# -*- coding: utf-8 -*-
"""
Automatic tests for module ldap0.filter
"""

# from Python's standard lib
import time
import unittest
from collections import OrderedDict

from ldap0.functions import strf_secs, escape_format
from ldap0.filter import (
    escape_str,
    escape_bytes,
    negate_filter,
    map_filter_parts,
    compose_filter,
    time_span_filter,
    dict_filter,
)


class TestFilter(unittest.TestCase):
    """
    test ldap0.filter
    """

    def test000_escape_str(self):
        """
        test function ldap0.filter.escape_str()
        """
        self.assertEqual(escape_str('foobar'), 'foobar')
        self.assertEqual(escape_str('foo\\bar'), 'foo\\5cbar')
        self.assertEqual(escape_str('foo*bar'), 'foo\\2abar')
        self.assertEqual(escape_str('Täster'), 'Täster')

    def test001_escape_bytes(self):
        """
        test function ldap0.filter.escape_bytes()
        """
        self.assertEqual(
            escape_bytes(b'\xc3\xa4\xc3\xb6\xc3\xbc\xc3\x84\xc3\x96\xc3\x9c\xc3\x9f'),
            '\\c3\\a4\\c3\\b6\\c3\\bc\\c3\\84\\c3\\96\\c3\\9c\\c3\\9f'
        )

    def test004_negate_filter(self):
        self.assertEqual(negate_filter('(objectClass=*)'), '(!(objectClass=*))')
        self.assertEqual(negate_filter('(!(objectClass=*))'), '(objectClass=*)')
        self.assertEqual(negate_filter('(!(!(objectClass=*)))'), '(!(objectClass=*))')

    def test005_map_filter_parts(self):
        self.assertEqual(map_filter_parts('foo', []), [])
        self.assertEqual(
            map_filter_parts('foo', ['1', '2']),
            ['(foo=1)', '(foo=2)']
        )
        self.assertEqual(
            map_filter_parts('sn', ['Täster', 'Müller']),
            ['(sn=Täster)', '(sn=Müller)']
        )

    def test007_compose_filter(self):
        self.assertEqual(compose_filter('', []), '')
        self.assertEqual(compose_filter('&', []), '')
        self.assertEqual(compose_filter('|', []), '')
        self.assertEqual(
            compose_filter('|', map_filter_parts('foo', ['1'])),
            '(foo=1)'
        )
        self.assertEqual(
            compose_filter('|', map_filter_parts('foo', ['1', '2'])),
            '(|(foo=1)(foo=2))'
        )

    def test008_dict_filter(self):
        self.assertEqual(dict_filter({}), '')
        self.assertEqual(dict_filter({'foo': ['bar']}), '(foo=bar)')
        self.assertEqual(
            dict_filter({'foo': ['bar1', 'bar2']}),
            '(&(foo=bar1)(foo=bar2))'
        )
        self.assertEqual(
            dict_filter({'foo': ['bar1', 'bar2']}, iop='|'), '(|(foo=bar1)(foo=bar2))'
        )
        self.assertEqual(
            dict_filter(
                OrderedDict([('foo1', ['bar1', 'bar2']), ('foo2', ['bar1', 'bar2'])]),
                iop='|'
            ),
            '(&(|(foo1=bar1)(foo1=bar2))(|(foo2=bar1)(foo2=bar2)))'
        )
        self.assertEqual(
            dict_filter(
                OrderedDict([('foo1', ['bar1', 'bar2']), ('foo2', ['bar1', 'bar2'])]),
                oop='|'
            ),
            '(|(&(foo1=bar1)(foo1=bar2))(&(foo2=bar1)(foo2=bar2)))'
        )
        self.assertEqual(
            dict_filter(
                OrderedDict([('foo1', ['T\xc3\xa4ster', 'M\xc3\xbcller']), ('foo2', ['T\xc3\xa4ster', 'M\xc3\xbcller'])]),
                oop='|'
            ),
            '(|(&(foo1=T\xc3\xa4ster)(foo1=M\xc3\xbcller))(&(foo2=T\xc3\xa4ster)(foo2=M\xc3\xbcller)))'
        )

    def test009_escape_format_with_filter(self):
        self.assertEqual(escape_format(escape_str, ''), '')
        self.assertEqual(escape_format(escape_str, '', 'foo', 'bar'), '')
        self.assertEqual(escape_format(escape_str, '({}={})', 'foo', 'bar'), '(foo=bar)')
        self.assertEqual(escape_format(escape_str, '(|(cn={})(cn={}))', 'foo', 'bar'), '(|(cn=foo)(cn=bar))')
        self.assertEqual(escape_format(escape_str, '(|(cn={})(cn={}))', 'fo*o', 'ba\\r'), '(|(cn=fo\\2ao)(cn=ba\\5cr))')
        self.assertEqual(escape_format(escape_str, '(|(cn={arg1})(cn={arg2}))', arg1='foo', arg2='bar'), '(|(cn=foo)(cn=bar))')
        self.assertEqual(escape_format(escape_str, '(|(cn={arg1})(cn={arg2}))', arg1='fo*o', arg2='ba\\r'), '(|(cn=fo\\2ao)(cn=ba\\5cr))')
        self.assertEqual(escape_format(escape_str, '(|({0}={arg1})({1}={arg2}))', 'foo', 'bar', arg1='fo*o', arg2='ba\\r'), '(|(foo=fo\\2ao)(bar=ba\\5cr))')

    def test010_time_span_filter(self):
        self.assertEqual(
            time_span_filter(until_timestamp=0),
            '(&(modifyTimestamp>=19700101000000Z)(!(modifyTimestamp>=19700101000000Z)))',
        )
        self.assertEqual(
            time_span_filter(
                from_timestamp=1000000,
                until_timestamp=1546101638,
                delta_attr='foo',
            ),
            '(&(foo>=19700112134640Z)(!(foo>=20181229164038Z)))',
        )
        with self.assertRaises(ValueError):
            time_span_filter(from_timestamp=1, until_timestamp=0)
        # test time.time() based filter
        before = time.time()
        flt = time_span_filter(delta_attr='foo')
        after = time.time()
        results = set([
            '(&(foo>=19700101000000Z)(!(foo>=%s)))' % (strf_secs(before)),
            '(&(foo>=19700101000000Z)(!(foo>=%s)))' % (strf_secs(after)),
        ])
        self.assertIn(flt, results)
        # test time.time() based filter with relative from_timestamp
        fts = -10
        before = time.time()
        flt = time_span_filter(from_timestamp=fts, delta_attr='foo')
        after = time.time()
        results = set([
            '(&(foo>=%s)(!(foo>=%s)))' % (strf_secs(before+fts), strf_secs(before)),
            '(&(foo>=%s)(!(foo>=%s)))' % (strf_secs(after+fts), strf_secs(after)),
        ])
        self.assertIn(flt, results)


if __name__ == '__main__':
    unittest.main()



# -*- coding: ascii -*-
"""
ldap0.controls.psearch - classes for Persistent Search Control
(see https://tools.ietf.org/html/draft-ietf-ldapext-psearch)
"""

from typing import Optional

# Imports from pyasn1
from pyasn1.type import namedtype, namedval, univ, constraint
from pyasn1.codec.ber import encoder, decoder
from pyasn1_modules.rfc2251 import LDAPDN

from . import RequestControl, ResponseControl

__all__ = [
    'PersistentSearchControl',
    'EntryChangeNotificationControl',
    'CHANGE_TYPES_INT',
    'CHANGE_TYPES_STR',
]

#---------------------------------------------------------------------------
# Constants and classes for Persistent Search Control
#---------------------------------------------------------------------------

CHANGE_TYPES_INT = {
    'add': 1,
    'delete': 2,
    'modify': 4,
    'modDN': 8,
}
CHANGE_TYPES_STR = {
    (v, k) for k, v in CHANGE_TYPES_INT.items()
}


class PersistentSearchControl(RequestControl):
    """
    Implements the request control for persistent search.

    changeTypes
      List of strings specifiying the types of changes returned by the server.
      Setting to None requests all changes.
    changesOnly
      Boolean which indicates whether only changes are returned by the server.
    returnECs
      Boolean which indicates whether the server should return an
      Entry Change Notication response control
    """
    controlType: str = '2.16.840.1.113730.3.4.3'

    class PersistentSearchControlValue(univ.Sequence):
        """
        PersistentSearch ::= SEQUENCE {
                changeTypes INTEGER,
                changesOnly BOOLEAN,
                returnECs BOOLEAN
        }
        """
        componentType = namedtype.NamedTypes(
            namedtype.NamedType('changeTypes', univ.Integer()),
            namedtype.NamedType('changesOnly', univ.Boolean()),
            namedtype.NamedType('returnECs', univ.Boolean()),
        )

    def __init__(
            self,
            criticality: bool = True,
            changeTypes: Optional[int] = None,
            changesOnly: bool = False,
            returnECs: bool = True,
        ):
        RequestControl.__init__(self, criticality=criticality)
        self.changesOnly = changesOnly
        self.returnECs = returnECs
        if isinstance(changeTypes, int):
            self.changeTypes = changeTypes
        else:
            # Assume a sequence type of integers to be OR-ed
            self.changeTypes = 0
            for chtype in changeTypes or CHANGE_TYPES_INT.values():
                self.changeTypes = self.changeTypes|CHANGE_TYPES_INT.get(chtype, chtype)
        # end of PersistentSearchControl()

    def encode(self) -> bytes:
        pval = self.PersistentSearchControlValue()
        pval.setComponentByName('changeTypes', univ.Integer(self.changeTypes))
        pval.setComponentByName('changesOnly', univ.Boolean(self.changesOnly))
        pval.setComponentByName('returnECs', univ.Boolean(self.returnECs))
        return encoder.encode(pval)


class ChangeType(univ.Enumerated):
    """
    changeType ENUMERATED {
        add             (1),
        delete          (2),
        modify          (4),
        modDN           (8)
    }
    """
    namedValues = namedval.NamedValues(
        ('add', 1),
        ('delete', 2),
        ('modify', 4),
        ('modDN', 8),
    )
    subtypeSpec = univ.Enumerated.subtypeSpec + constraint.SingleValueConstraint(1, 2, 4, 8)


class EntryChangeNotificationValue(univ.Sequence):
    """
    EntryChangeNotification ::= SEQUENCE {
        changeType <see ChangeType above>,
        previousDN   LDAPDN OPTIONAL,     -- modifyDN ops. only
        changeNumber INTEGER OPTIONAL     -- if supported
    }
    """
    componentType = namedtype.NamedTypes(
        namedtype.NamedType('changeType', ChangeType()),
        namedtype.OptionalNamedType('previousDN', LDAPDN()),
        namedtype.OptionalNamedType('changeNumber', univ.Integer()),
    )


class EntryChangeNotificationControl(ResponseControl):
    """
    Implements the response control for persistent search.

    Class attributes with values extracted from the response control:

    changeType
      String indicating the type of change causing this result to be
      returned by the server
    previousDN
      Old DN of the entry in case of a modrdn change
    changeNumber
      A change serial number returned by the server (optional).
    """
    controlType: str = '2.16.840.1.113730.3.4.7'

    def decode(self, encodedControlValue: bytes):
        ecncValue, _ = decoder.decode(
            encodedControlValue,
            asn1Spec=EntryChangeNotificationValue(),
        )
        self.changeType = int(ecncValue.getComponentByName('changeType'))
        previousDN = ecncValue.getComponentByName('previousDN')
        if previousDN.hasValue():
            self.previousDN = bytes(previousDN)
        else:
            self.previousDN = None
        changeNumber = ecncValue.getComponentByName('changeNumber')
        if changeNumber.hasValue():
            self.changeNumber = int(changeNumber)
        else:
            self.changeNumber = None
        return (self.changeType, self.previousDN, self.changeNumber)



# -*- coding: ascii -*-
"""
ldap0.err - error classes
"""

import _libldap0
from . import LDAPError

__all__ = [
    'LIMIT_EXCEPTIONS',
    'NoUniqueEntry',
    'PasswordPolicyException',
    'PasswordPolicyChangeAfterReset',
    'PasswordPolicyExpirationWarning',
    'PasswordPolicyExpiredError',
]

LIMIT_EXCEPTIONS = (
    _libldap0.TIMEOUT,
    _libldap0.TIMELIMIT_EXCEEDED,
    _libldap0.SIZELIMIT_EXCEEDED,
    _libldap0.ADMINLIMIT_EXCEEDED,
)

#-----------------------------------------------------------------------
# Some custom exception classes for password policy handling
#-----------------------------------------------------------------------

class PasswordPolicyException(LDAPError):
    """
    Base class for raising password policy related exceptions
    """

    def __init__(self, who=None, desc=None):
        self.who = who
        self.desc = desc

    def __str__(self):
        return self.desc


class PasswordPolicyChangeAfterReset(PasswordPolicyException):
    """
    Exception class for password change after reset warning
    """


class PasswordPolicyExpirationWarning(PasswordPolicyException):
    """
    Exception class for password expiry warning
    """

    def __init__(self, who=None, desc=None, timeBeforeExpiration=None):
        PasswordPolicyException.__init__(self, who, desc)
        self.timeBeforeExpiration = timeBeforeExpiration


class PasswordPolicyExpiredError(PasswordPolicyException):
    """
    Exception class for password expired error
    """

    def __init__(self, who=None, desc=None, graceAuthNsRemaining=None):
        PasswordPolicyException.__init__(self, who, desc)
        self.graceAuthNsRemaining = graceAuthNsRemaining


#-----------------------------------------------------------------------
# Some custom exception classes for search errors
#-----------------------------------------------------------------------

class NoUniqueEntry(_libldap0.NO_SUCH_OBJECT):
    """
    Exception raised if a LDAP search returned more than entry entry
    although assumed to return a unique single search result.
    """



# -*- coding: utf-8 -*-
"""
Simple tests for module ldap0.cache
"""

# from Python's standard lib
import time
import unittest

import ldap0.cache


class TestCache(unittest.TestCase):
    """
    test ldap0.cache.Cache
    """

    def test_cache_simple(self):
        cache = ldap0.cache.Cache(ttl=1.0)
        with self.assertRaises(KeyError):
            cache['foo']
        self.assertEqual(cache.miss_count, 1)
        # seed the cache
        cache['foo'] = 'bar'
        self.assertEqual(len(cache), 1)
        orig_miss_count = cache.miss_count
        for i in range(0, 5):
            self.assertEqual(cache['foo'], 'bar')
            self.assertEqual(cache.hit_count, i+1)
            self.assertEqual(cache.miss_count, orig_miss_count)
        self.assertEqual(cache.hit_count, 5)
        # test expiry
        time.sleep(1.1)
        with self.assertRaises(KeyError):
            cache['foo']
        self.assertEqual(len(cache), 0)

    def test_cache_flush(self):
        cache = ldap0.cache.Cache(ttl=1.0)
        # seed the cache
        cache['foo'] = 'bar'
        self.assertEqual(len(cache), 1)
        self.assertEqual(cache['foo'], 'bar')
        self.assertEqual(cache.miss_count, 0)
        # test whether flushing cache works
        cache.flush()
        self.assertEqual(cache.hit_count, 0)
        self.assertEqual(cache.miss_count, 0)
        self.assertEqual(len(cache), 0)
        with self.assertRaises(KeyError):
            cache['foo']


if __name__ == '__main__':
    unittest.main()



# -*- coding: ascii -*-
"""
ldap0.test - module for spawning test instances of OpenLDAP's slapd server
"""

import os
import socket
import time
import subprocess
import logging
import unittest
from typing import List

from urllib.parse import quote_plus

import _libldap0

from .base import encode_list
from .ldapobject import LDAPObject
from .functions import escape_format
from .filter import escape_str as escape_filter_str
from .ldif import LDIFParser
from .typehints import EntryStr

# a template string for generating simple slapd.conf file
SLAPD_CONF_TEMPLATE = r"""
moduleload back_%(database)s

include "%(schema_prefix)s/core.schema"
include "%(schema_prefix)s/cosine.schema"
include "%(schema_prefix)s/nis.schema"

serverID %(serverid)s
loglevel %(loglevel)s
allow bind_v2
sasl-secprops none
security ssf=0 simple_bind=0 sasl=0 transport=0

TLSVerifyClient allow
TLSProtocolMin 3.3
TLSCACertificateFile tests/tls/ca-chain.pem
TLSCertificateFile tests/tls/localhost.crt
TLSCertificateKeyFile tests/tls/localhost.key
TLSDHParamFile tests/tls/slapd-dhparams.pem

authz-regexp
  "gidnumber=%(root_gid)s\\+uidnumber=%(root_uid)s,cn=peercred,cn=external,cn=auth"
  "%(rootdn)s"

authz-regexp
  "uid=([a-zA-Z0-9_-]+),cn=(DIGEST-MD5|CRAM-MD5|PLAIN|SCRAM-SHA-1),cn=auth"
  "ldap:///%(suffix)s??sub?(cn=$1)"

database config
access to
  dn.subtree="cn=config"
    by dn.exact="%(rootdn)s" manage

database %(database)s
directory "%(directory)s"
suffix "%(suffix)s"
rootdn "%(rootdn)s"
rootpw "%(rootpw)s"

dbnosync

access to
  dn.subtree="%(suffix)s"
  attrs=userPassword
    by self =w
    by anonymous auth

access to
  dn.subtree="%(suffix)s"
    by * auth
"""

LOCALHOST = '127.0.0.1'

CONSOLE_LOG_FORMAT = '%(name)s %(asctime)s %(levelname)s %(message)s'


class SlapdObject:
    """
    Controller class for a slapd instance, OpenLDAP's server.

    This class creates a temporary data store for slapd, runs it
    listening on a private Unix domain socket and TCP port,
    and initialises it with a top-level entry and the root user.

    When a reference to an instance of this class is lost, the slapd
    server is shut down.
    """
    slapd_conf_template = SLAPD_CONF_TEMPLATE
    database = 'mdb'
    suffix = 'o=\xE4\xF6\xFC,dc=example,dc=org'
    root_cn = 'M\xE4nager'
    root_dn = 'cn=%s,%s' % (root_cn, suffix)
    root_pw = 'password'
    slapd_loglevel = [
        v.strip()
        for v in os.environ.get('SLAPD_LOGLEVEL', 'stats,stats2').split(',')
    ]
    # use SASL/EXTERNAL via LDAPI when invoking OpenLDAP CLI tools
    cli_sasl_external = True
    local_host = '127.0.0.1'
    testrunsubdirs = (
        'schema',
    )
    openldap_schema_files = (
        'core.schema',
        'cosine.schema',
        'nis.schema',
    )

    TMPDIR = os.environ.get('TMP', os.getcwd())
    BINDIR = os.environ.get('BIN', '/usr/bin')
    SCHEMADIR = os.environ.get('SCHEMA', '/etc/openldap/schema')
    PATH_LDAPADD = os.path.join(BINDIR, 'ldapadd')
    PATH_LDAPMODIFY = os.path.join(BINDIR, 'ldapmodify')
    PATH_LDAPWHOAMI = os.path.join(BINDIR, 'ldapwhoami')
    PATH_SLAPD = os.environ.get('SLAPD', '/usr/sbin/slapd')

    # time in secs to wait before trying to access slapd via LDAP (again)
    start_sleep = float(os.environ.get('START_SLEEP', '1.5'))

    def __init__(self):
        self._proc = None
        self._port = self._avail_tcp_port()
        self.server_id = self._port % 4096
        self.testrundir = os.path.join(self.TMPDIR, 'slapdtest-%d' % self._port)
        self._schema_prefix = os.path.join(self.testrundir, 'schema')
        self._slapd_conf = os.path.join(self.testrundir, 'slapd.conf')
        self._db_directory = os.path.join(self.testrundir, "openldap-data")
        self.ldap_uri = "ldap://%s:%d/" % (LOCALHOST, self._port)
        ldapi_path = os.path.join(self.testrundir, 'ldapi')
        self.ldapi_uri = "ldapi://%s" % quote_plus(ldapi_path)
        self._slapd_version = None

    @property
    def slapd_version(self):
        if self._slapd_version is None:
            self._slapd_version = '.'.join(self.get_slapd_version())
        return self._slapd_version

    def setup_rundir(self):
        """
        creates rundir structure

        for setting up a custom directory structure you have to override
        this method
        """
        os.mkdir(self.testrundir)
        os.mkdir(self._db_directory)
        self._create_sub_dirs(self.testrunsubdirs)
        self._ln_schema_files(self.openldap_schema_files, self.SCHEMADIR)

    def _cleanup_rundir(self):
        """
        Recursively delete whole directory specified by `path'
        """
        if not os.path.exists(self.testrundir):
            return
        logging.debug('clean-up %s', self.testrundir)
        for dirpath, dirnames, filenames in os.walk(
                self.testrundir,
                topdown=False
            ):
            for filename in filenames:
                logging.debug('remove %s', os.path.join(dirpath, filename))
                os.remove(os.path.join(dirpath, filename))
            for dirname in dirnames:
                logging.debug('rmdir %s', os.path.join(dirpath, dirname))
                os.rmdir(os.path.join(dirpath, dirname))
        os.rmdir(self.testrundir)
        logging.info('cleaned-up %s', self.testrundir)

    def _avail_tcp_port(self) -> int:
        """
        find an available port for TCP connection
        """
        sock = socket.socket()
        sock.bind((self.local_host, 0))
        port = sock.getsockname()[1]
        sock.close()
        logging.info('Found available port %d', port)
        return port

    def gen_config(self) -> str:
        """
        generates a slapd.conf and returns it as one string

        for generating specific static configuration files you have to
        override this method
        """
        config_dict = {
            'serverid': hex(self.server_id),
            'schema_prefix':self._schema_prefix,
            'loglevel': ' '.join(self.slapd_loglevel),
            'database': self.database,
            'directory': self._db_directory,
            'suffix': self.suffix,
            'rootdn': self.root_dn,
            'rootpw': self.root_pw,
            'root_uid': os.getuid(),
            'root_gid': os.getgid(),
        }
        return self.slapd_conf_template % config_dict

    def _create_sub_dirs(self, dir_names):
        """
        create sub-directories beneath self.testrundir
        """
        for dname in dir_names:
            dir_name = os.path.join(self.testrundir, dname)
            logging.debug('Create directory %s', dir_name)
            os.mkdir(dir_name)

    def _ln_schema_files(self, file_names, source_dir):
        """
        write symbolic links to original schema files
        """
        for fname in file_names:
            ln_source = os.path.join(source_dir, fname)
            ln_target = os.path.join(self._schema_prefix, fname)
            logging.debug('Create symlink %s -> %s', ln_source, ln_target)
            os.symlink(ln_source, ln_target)

    def _write_config(self):
        """Writes the slapd.conf file out, and returns the path to it."""
        logging.debug('Writing config to %s', self._slapd_conf)
        with open(self._slapd_conf, 'w', encoding='utf-8') as config_file:
            config_file.write(self.gen_config())
        logging.info('Wrote config to %s', self._slapd_conf)

    def _test_config(self):
        logging.debug('testing config %s', self._slapd_conf)
        popen_list = [
            self.PATH_SLAPD,
            '-T', 'test',
            '-f', self._slapd_conf,
            '-u',
        ]
        if logging.getLogger().isEnabledFor(logging.DEBUG):
            popen_list.append('-v')
            popen_list.extend(['-d', 'config'])
        else:
            popen_list.append('-Q')
        proc = subprocess.Popen(popen_list)
        if proc.wait() != 0:
            raise RuntimeError("configuration test failed")
        logging.info("config ok: %s", self._slapd_conf)

    def get_slapd_version(self) -> tuple:
        """
        Determines slapd version
        """
        proc = subprocess.Popen(
            [self.PATH_SLAPD, '-V'],
            stdin=subprocess.PIPE,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
        )
        _, stderr_data = proc.communicate(None)
        return tuple(stderr_data.decode('utf-8').split('\n')[0].split(' ')[3].split('.'))

    def _start_slapd(self):
        """
        Spawns/forks the slapd process
        """
        slapd_args = [
            self.PATH_SLAPD,
            '-f', self._slapd_conf,
            '-h', '%s' % ' '.join((self.ldap_uri, self.ldapi_uri)),
        ]
        if logging.getLogger().isEnabledFor(logging.DEBUG):
            slapd_args.extend(['-d', ','.join(self.slapd_loglevel)])
        else:
            slapd_args.extend(['-d', '0'])
        logging.info(
            'starting slapd %s: %r',
            '.'.join(self.get_slapd_version()),
            ' '.join(slapd_args),
        )
        self._proc = subprocess.Popen(slapd_args)
        # Waits until the LDAP server socket is open, or slapd crashed
        while True:
            if self._proc.poll() is not None:
                self._stopped()
                raise RuntimeError("slapd exited before opening port")
            time.sleep(self.start_sleep)
            try:
                logging.debug("slapd connection check to %s", self.ldapi_uri)
                self.ldapwhoami()
            except RuntimeError:
                pass
            else:
                return

    def start(self):
        """
        Starts the slapd server process running, and waits for it to come up.
        """

        if self._proc is None:
            # prepare directory structure
            self._cleanup_rundir()
            self.setup_rundir()
            self._write_config()
            self._test_config()
            self._start_slapd()
            logging.debug(
                'slapd with pid=%d listening on %s and %s',
                self._proc.pid, self.ldap_uri, self.ldapi_uri
            )

    def stop(self):
        """
        Stops the slapd server, and waits for it to terminate and cleans up
        """
        if self._proc is not None:
            logging.debug('stopping slapd with pid %d', self._proc.pid)
            self._proc.terminate()
            self.wait()
        self._cleanup_rundir()

    def restart(self):
        """
        Restarts the slapd server with same data
        """
        self._proc.terminate()
        self.wait()
        self._start_slapd()

    def wait(self):
        """Waits for the slapd process to terminate by itself."""
        if self._proc:
            self._proc.wait()
            self._stopped()

    def _stopped(self):
        """Called when the slapd server is known to have terminated"""
        if self._proc is not None:
            logging.info('slapd[%d] terminated', self._proc.pid)
            self._proc = None

    def _cli_auth_args(self) -> List[str]:
        if self.cli_sasl_external:
            authc_args = [
                '-Y', 'EXTERNAL',
            ]
            if not logging.getLogger().isEnabledFor(logging.DEBUG):
                authc_args.append('-Q')
        else:
            authc_args = [
                '-x',
                '-D', self.root_dn,
                '-w', self.root_pw,
            ]
        return authc_args

    def _cli_popen(self, ldapcommand, extra_args=None, ldap_uri=None, stdin_data=None):
        args = [
            ldapcommand,
            '-H', ldap_uri or self.ldapi_uri,
        ] + self._cli_auth_args() + (extra_args or [])
        logging.debug('Run command: %r', ' '.join(args))
        proc = subprocess.Popen(args, stdin=subprocess.PIPE, stdout=subprocess.PIPE)
        logging.debug('stdin_data=%r', stdin_data)
        stdout_data, stderr_data = proc.communicate(stdin_data)
        if stdout_data is not None:
            logging.debug('stdout_data=%r', stdout_data)
        if stderr_data is not None:
            logging.debug('stderr_data=%r', stderr_data)
        if proc.wait() != 0:
            raise RuntimeError('ldapadd process failed')
        return stdout_data, stderr_data

    def ldapwhoami(self, extra_args=None):
        """
        Runs ldapwhoami on this slapd instance
        """
        self._cli_popen(self.PATH_LDAPWHOAMI, extra_args=extra_args)

    def ldapadd(self, ldif_data: bytes, extra_args=None):
        """
        Runs ldapadd on this slapd instance, passing it the ldif content
        """
        self._cli_popen(self.PATH_LDAPADD, extra_args=extra_args, stdin_data=ldif_data)

    def ldapmodify(self, ldif_data: bytes, extra_args=None):
        """
        Runs ldapadd on this slapd instance, passing it the ldif content
        """
        self._cli_popen(self.PATH_LDAPMODIFY, extra_args=extra_args, stdin_data=ldif_data)


class SlapdTestCase(unittest.TestCase):
    """
    test class which also clones or initializes a running slapd
    """

    server_class = SlapdObject
    ldap_object_class = LDAPObject
    server = None
    trace_level = int(os.environ.get('LDAP0_TRACE_LEVEL', '0'))

    def _open_ldap_conn(self, who=None, cred=None):
        """
        return a LDAPObject instance after simple bind
        """
        ldap_conn = self.ldap_object_class(
            self.server.ldap_uri,
            trace_level=self.trace_level,
        )
        ldap_conn.protocol_version = 3
        #ldap_conn.set_option(ldap0.OPT_REFERRALS, 0)
        ldap_conn.simple_bind_s(
            (who or self.server.root_dn),
            (cred or self.server.root_pw).encode('utf-8'),
        )
        return ldap_conn

    @classmethod
    def setUpClass(cls):
        logging.getLogger().setLevel(os.environ.get('LDAP0_LOG_LEVEL', 'WARN').upper())
        cls.server = cls.server_class()
        cls.server.start()

    @classmethod
    def tearDownClass(cls):
        cls.server.stop()

    def setUp(self):
        try:
            self._ldap_conn
        except AttributeError:
            # open local LDAP connection
            self._ldap_conn = self._open_ldap_conn()
            self.mdb_config_dn = self._ldap_conn.search_s(
                'cn=config',
                _libldap0.SCOPE_SUBTREE,
                escape_format(escape_filter_str, '(olcSuffix={0})', self.server.suffix),
            )[0].dn_s

    def enable_overlay(self, slapo_entry: EntryStr, schema_files=None, config_dn=None):
        logging.debug('Enable overlay with %r', slapo_entry)
        olc_overlay = slapo_entry['olcOverlay'][0]
        try:
            enabled_overlays = [
                val.split('}', 1)[1]
                for val in self._ldap_conn.read_s(
                    'cn=module{0},cn=config',
                    attrlist=['olcModuleLoad'],
                ).entry_s['olcModuleLoad']
            ]
            if olc_overlay in enabled_overlays:
                return
        except _libldap0.NO_SUCH_OBJECT:
            enabled_overlays = []

        schema_entries = []
        for schema_filename in schema_files or []:
            with open(os.path.join(self.server.SCHEMADIR, schema_filename), 'rb') as schema_file:
                ldif_parser = LDIFParser(schema_file)
                schema_entries.extend(ldif_parser.parse())
        try:
            for dn, entry in schema_entries:
                self._ldap_conn.add_s(
                    dn.decode('utf-8'),
                    {at.decode('ascii'): avs for at, avs in entry.items()}
                )
            if enabled_overlays:
                self._ldap_conn.modify_s(
                    'cn=module{0},cn=config',
                    [(_libldap0.MOD_ADD, b'olcModuleLoad', [olc_overlay.encode('utf-8')])]
                )
            else:
                self._ldap_conn.add_s(
                    'cn=module{0},cn=config',
                    {
                        'objectclass': ['olcModuleList'.encode('utf-8')],
                        'olcModuleload': [olc_overlay.encode('utf-8')],
                    }
                )
            self._ldap_conn.add_s(
                'olcOverlay={{0}}{0},{1}'.format(
                    olc_overlay,
                    config_dn or self.mdb_config_dn,
                ),
                {
                    at: encode_list(avs)
                    for at, avs in slapo_entry.items()
                },
            )
        except _libldap0.OTHER as ldap_err:
            self.skipTest(
                'installing slapo-%s failed: %s' % (
                    olc_overlay,
                    ldap_err,
                )
            )



# -*- coding: utf-8 -*-
"""
Automatic tests for module ldap0.dn
"""

# from Python's standard lib
import hashlib
import unittest

from ldap0.pw import *


class TestPasswordFunctions(unittest.TestCase):
    """
    test functions in ldap0.dn
    """

    def test_random_string(self):
        self.assertEqual(len(random_string()), PWD_LENGTH)
        self.assertEqual(random_string(b'1', 4), 4*b'1')
        self.assertEqual(random_string('1', 4), 4*'1')
        self.assertEqual(len(random_string(None)), PWD_LENGTH)

    def test_unicode_pwd(self):
        self.assertEqual(len(unicode_pwd()), (PWD_LENGTH+2)*2)
        self.assertEqual(len(unicode_pwd(length=10)), (10+2)*2)
        for pw in (
                b'foo',
                b'bar',
                b'\xc3\xa4\xc3\xb6\xc3\xbc\xc3\x84\xc3\x96\xc3\x9c\xc3\x9f',
            ):
            self.assertEqual(
                unicode_pwd(pw),
                '"{}"'.format(pw.decode('utf-8')).encode('utf-16-le')
            )

    def test_ntlm_password_hash(self):
        ntlm_password_hash_hex_size = 2 * hashlib.new('md4').digest_size
        for pw, pw_hash in (
                (b'foo', b'AC8E657F83DF82BEEA5D43BDAF7800CC'),
                (b'bar', b'86C156FC198B358CCCF6278D8BD49B6A'),
                (b'\xc3\xa4\xc3\xb6\xc3\xbc\xc3\x84\xc3\x96\xc3\x9c\xc3\x9f', b'F4B08969FAAC3D4A52924FA91944261E'),
            ):
            res = ntlm_password_hash(pw)
            self.assertEqual(len(res), ntlm_password_hash_hex_size)
            self.assertEqual(res, pw_hash)


if __name__ == '__main__':
    unittest.main()



# -*- coding: ascii -*-
"""
ldap0.controls.pagedresults - classes for Simple Paged control
(see RFC 2696)
"""

# Imports from pyasn1
from pyasn1.type import namedtype, univ
from pyasn1.codec.ber import encoder, decoder
from pyasn1_modules.rfc2251 import LDAPString

from . import RequestControl, ResponseControl


__all__ = [
    'SimplePagedResultsControl'
]

class SimplePagedResultsControl(RequestControl, ResponseControl):
    controlType: str = '1.2.840.113556.1.4.319'

    class PagedResultsControlValue(univ.Sequence):
        componentType = namedtype.NamedTypes(
            namedtype.NamedType('size', univ.Integer()),
            namedtype.NamedType('cookie', LDAPString()),
        )

    def __init__(self, criticality: bool = False, size=10, cookie=''):
        RequestControl.__init__(self, criticality=criticality)
        self.size = size
        self.cookie = cookie or b''

    def encode(self) -> bytes:
        pc = self.PagedResultsControlValue()
        pc.setComponentByName('size', univ.Integer(self.size))
        pc.setComponentByName('cookie', LDAPString(self.cookie))
        return encoder.encode(pc)

    def decode(self, encodedControlValue: bytes):
        decodedValue, _ = decoder.decode(
            encodedControlValue,
            asn1Spec=self.PagedResultsControlValue(),
        )
        self.size = int(decodedValue.getComponentByName('size'))
        self.cookie = decodedValue.getComponentByName('cookie')



# -*- coding: ascii -*-
"""
ldap0.msad - proprietary stuff for MS Actiev Directory
"""

import struct
import uuid

__all__ = [
    'sid2sddl',
    'sddl2sid',
    'ObjectGUID',
]


def sddl2sid(sddl: str) -> bytes:
    """
    convert SDDL str representation of a SID to bytes

    implements the inverse of function sid2sddl()
    """
    sid_components = sddl.split('-')
    srl_byte = int(sid_components[1])
    number_sub_id_byte = len(sid_components)-3
    result_list = [bytes([srl_byte, number_sub_id_byte])]
    iav_buf = struct.pack('!Q', int(sid_components[2]))[2:]
    result_list.append(iav_buf)
    result_list.extend([
        struct.pack('<I', int(s))
        for s in sid_components[3:]
    ])
    return b''.join(result_list)


def sid2sddl(sid: bytes) -> str:
    """
    convert SID bytes to SDDL str representation

    implements the inverse of sddl2sid()
    """
    srl = sid[0]
    number_sub_id = sid[1]
    iav = struct.unpack('!Q', b'\x00\x00'+sid[2:8])[0]
    sub_ids = [
        struct.unpack('<I', sid[8+4*i:12+4*i])[0]
        for i in range(number_sub_id)
    ]
    return 'S-%d-%d-%s' % (
        srl,
        iav,
        '-'.join([str(s) for s in sub_ids]),
    )



# -*- coding: ascii -*-
"""
ldap0.controls.libldap - LDAP controls wrapper classes with en-/decoding done
by OpenLDAP functions
"""

import _libldap0

from . import RequestControl


class AssertionControl(RequestControl):
    """
    LDAP Assertion control, as defined in RFC 4528

    filterstr
      LDAP filter string specifying which assertions have to match
      so that the server processes the operation
    """
    controlType: str = '1.3.6.1.1.12'

    def __init__(self, criticality: bool = True, filterstr: str = '(objectClass=*)'):
        RequestControl.__init__(self, criticality=criticality)
        self.filterstr = filterstr

    def encode(self) -> bytes:
        return _libldap0.encode_assertion_control(self.filterstr.encode(self.encoding))


class MatchedValuesControl(RequestControl):
    """
    LDAP Matched Values control, as defined in RFC 3876

    filterstr
      LDAP filter string specifying which attribute values
      should be returned
    """
    controlType: str = '1.2.826.0.1.3344810.2.3'

    def __init__(self, criticality: bool = False, filterstr: str = '(objectClass=*)'):
        RequestControl.__init__(self, criticality=criticality)
        self.filterstr = filterstr

    def encode(self) -> bytes:
        return _libldap0.encode_valuesreturnfilter_control(self.filterstr.encode(self.encoding))



# -*- coding: ascii -*-
"""
ldap0.functions - wraps functions of module _ldap
"""

from typing import Optional, Set, Union

__all__ = [
    'escape_str',
    'get_option',
    'set_option',
    'strf_secs',
    'strp_secs',
]

import datetime
import pprint
import time
import logging
from calendar import timegm

import _libldap0
from _libldap0 import LDAPError

from .lock import LDAPLock
import ldap0

global _LIBLDAP0_MODULE_LOCK
_LIBLDAP0_MODULE_LOCK = LDAPLock(desc='Module lock', trace_level=ldap0._trace_level)

# format string used for Generalized Time attributes values with UTC
DATETIME_FORMAT_UTC = r'%Y%m%d%H%M%SZ'


def _libldap0_function_call(func, *args, **kwargs):
    """
    Wrapper function which locks and logs calls to function

    lock
        Instance of threading.Lock or compatible
    func
        Function to call with arguments passed in via *args and **kwargs
    """
    if __debug__ and ldap0._trace_level >= 1:
        logging.debug(
            '_libldap0.%s %s\n',
            func.__name__,
            pprint.pformat((args, kwargs)),
            exc_info=(ldap0._trace_level >= 9),
        )
    _LIBLDAP0_MODULE_LOCK.acquire()
    try: # finally
        try: # error / result logging
            result = func(*args, **kwargs)
        except LDAPError as err:
            if __debug__ and ldap0._trace_level >= 2:
                result_log_msg = str(err)
            raise
        else:
            if __debug__ and ldap0._trace_level >= 2:
                result_log_msg = pprint.pformat(result)
    finally:
        _LIBLDAP0_MODULE_LOCK.release()
    if __debug__ and ldap0._trace_level >= 2:
        logging.debug('-> %s', result_log_msg)
    return result # end of _libldap0_function_call()


def get_option(option):
    """
    get_option(name) -> value

    Get the value of an LDAP global option.
    """
    return _libldap0_function_call(_libldap0.get_option, option)


def set_option(option, invalue):
    """
    set_option(name, value)

    Set the value of an LDAP global option.
    """
    return _libldap0_function_call(_libldap0.set_option, option, invalue)


def escape_str(escape_str_func, fmt: str, *args) -> str:
    """
    Applies escape_func() to all items of `args' and returns a string based
    on format string `val'.
    """
    escaped_args = map(escape_str_func, args)
    return fmt % tuple(escaped_args)


def escape_format(escape_str_func, fmt: str, *args, **kwargs) -> str:
    escaped_args = [
        escape_str_func(val)
        for val in args
    ]
    escaped_kwargs = dict([
        (key, escape_str_func(val))
        for key, val in kwargs.items()
    ])
    return fmt.format(*escaped_args, **escaped_kwargs)


def strf_secs(secs: float) -> str:
    """
    Convert seconds since epoch to a string compliant to LDAP syntax GeneralizedTime
    """
    return time.strftime(DATETIME_FORMAT_UTC, time.gmtime(secs))


def strp_secs(dt_str: str) -> float:
    """
    Convert LDAP syntax GeneralizedTime to seconds since epoch
    """
    return timegm(time.strptime(dt_str, DATETIME_FORMAT_UTC))


def str2datetime(time_str: str):
    """
    Parse GeneralizedTime (UTC) string to datetime.datetime instance
    """
    return datetime.datetime.strptime(time_str, DATETIME_FORMAT_UTC)


def datetime2str(time_dt) -> str:
    """
    Return datetime as string formatted as GeneralizedTime (UTC)
    """
    return datetime.datetime.strftime(time_dt, DATETIME_FORMAT_UTC)


def is_expired(
        start: Union[str, bytes, datetime.datetime],
        max_age: Union[str, bytes, int, float, datetime.timedelta],
        now: Optional[datetime.datetime] = None,
        disable_secs: Union[int, float] = 0,
    ) -> bool:
    """
    Check expiry of something based on start date and time
    and a timespan.

    Returns True if start + max_age > now
    """
    if isinstance(start, bytes):
        start = str2datetime(start.decode('utf-8'))
    elif isinstance(start, str):
        start = str2datetime(start)
    elif isinstance(start, (int, float)):
        start = datetime.datetime.fromtimestamp(start)
    if isinstance(max_age, (str, bytes, float)):
        max_age = datetime.timedelta(seconds=int(max_age))
    elif isinstance(max_age, int):
        max_age = datetime.timedelta(seconds=max_age)
    if max_age.seconds <= disable_secs or not start:
        # nothing to check
        return False
    return (start + max_age) < (now or datetime.datetime.utcnow())


def attr_set(attrs_str: str) -> Set[str]:
    """
    Splits string with space or comma-separated list of attribute names
    and returns a set with white-space stripped values
    """
    res = set()
    for attr in (attrs_str or '').strip().replace(' ', ',').split(','):
        attr = attr.strip()
        if attr:
            res.add(attr)
    return res



# -*- coding: ascii -*-
"""
ldap0.pw - stuff for passwords
"""

import hashlib
import string
import secrets
from binascii import hexlify
from typing import Union, Optional

__all__ = [
    'ntlm_password_hash',
    'random_string',
    'unicode_pwd',
    'PWD_ALPHABET',
    'PWD_LENGTH',
]

# default length for generated passwords
PWD_LENGTH = 40

# alphabet used for generated passwords
PWD_ALPHABET = (string.ascii_letters + string.digits).encode('ascii')

# Alphabet for encrypted passwords (see module crypt)
PWD_UNIX_CRYPT_ALPHABET = b'./0123456789abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ'


def random_string(
        alphabet: Union[str, bytes, None] = PWD_ALPHABET,
        length: int = PWD_LENGTH,
    ) -> Union[str, bytes]:
    """
    Returns a random password byte string consisting of `length' bytes of
    chars found in `alphabet'.
    """
    if alphabet is None:
        return secrets.token_bytes(length)
    res = []
    for _ in range(length):
        res.append(secrets.choice(alphabet))
    if isinstance(alphabet, bytes):
        return bytes(res)
    return ''.join(res)


def unicode_pwd(
        password: Optional[bytes] = None,
        alphabet: bytes = PWD_ALPHABET,
        length: int = PWD_LENGTH,
    ) -> bytes:
    """
    returns password or random generated password as properly encoded
    'unicodePwd' value for MS AD

    :password: must be UTF-8 encoded bytes

    see also:
    https://msdn.microsoft.com/en-us/library/cc223248.aspx
    https://support.microsoft.com/en-us/help/269190/how-to-change-a-windows-active-directory-and-lds-user-password-through-ldap
    """
    if password is None:
        password = random_string(alphabet=alphabet, length=length)
    return '"{}"'.format(password.decode('utf-8')).encode('utf-16-le')


def ntlm_password_hash(password: bytes) -> bytes:
    """
    returns MD4-based NT password hash for a password.

    :password: must be UTF-8 encoded bytes
    """
    md4_digest = hashlib.new('md4')
    md4_digest.update(password.decode('utf-8')[:128].encode('utf-16-le'))
    return hexlify(md4_digest.digest()).upper()



# -*- coding: ascii -*-
"""
ldap0.ldapurl - handling of LDAP URLs as described in RFC 4516
"""

from typing import List, Optional

from urllib.parse import (
    quote as url_quote,
    unquote as url_unquote,
)
from collections import UserDict

__all__ = [
    # constants
    'SEARCH_SCOPE',
    'SEARCH_SCOPE_STR',
    'LDAP_SCOPE_BASE',
    'LDAP_SCOPE_ONELEVEL',
    'LDAP_SCOPE_SUBTREE',
    # functions
    'is_ldapurl',
    # classes
    'LDAPUrlExtension',
    'LDAPUrlExtensions',
    'LDAPUrl'
]

LDAP_SCOPE_BASE = 0
LDAP_SCOPE_ONELEVEL = 1
LDAP_SCOPE_SUBTREE = 2
LDAP_SCOPE_SUBORDINATES = 3

LDAPURL_SCHEMES = {
    'ldap',
    'ldaps',
    'ldapi',
}

SEARCH_SCOPE_STR = {
    None: '',
    LDAP_SCOPE_BASE: 'base',
    LDAP_SCOPE_ONELEVEL: 'one',
    LDAP_SCOPE_SUBTREE: 'sub',
    LDAP_SCOPE_SUBORDINATES: 'subordinates',
}

SEARCH_SCOPE = {
    '': None,
    # the search scope strings defined in RFC 4516
    'base': LDAP_SCOPE_BASE,
    'one': LDAP_SCOPE_ONELEVEL,
    'sub': LDAP_SCOPE_SUBTREE,
    # from draft-sermersheim-ldap-subordinate-scope
    'subordinates': LDAP_SCOPE_SUBORDINATES,
}


def is_ldapurl(val):
    """
    Returns True if s is a LDAP URL, False else
    """
    try:
        scheme, _ = val.split('://', 1)
    except ValueError:
        return False
    return scheme.lower() in LDAPURL_SCHEMES


def escape_str(val: str) -> str:
    """
    Returns URL encoding of val
    """
    return url_quote(val, safe='', encoding='utf-8').replace('/', '%2F')


class LDAPUrlExtension:
    """
    Class for parsing and unparsing LDAP URL extensions
    as described in RFC 4516.

    Usable class attributes:
      critical
            Boolean integer marking the extension as critical
      extype
            Type of extension
      exvalue
            Value of extension
    """
    critical: bool
    extype: Optional[str]
    exvalue: Optional[str]

    def __init__(
            self,
            extensionStr: Optional[str] = None,
            critical: bool = False,
            extype: Optional[str] = None,
            exvalue: Optional[str] = None
        ):
        self.critical = critical
        self.extype = extype
        self.exvalue = exvalue
        if extensionStr:
            self._parse(extensionStr)

    def _parse(self, extension: str):
        extension = extension.strip()
        if not extension:
            # Don't parse empty strings
            self.extype, self.exvalue = None, None
            return
        self.critical = extension[0] == '!'
        if extension[0] == '!':
            extension = extension[1:].strip()
        try:
            self.extype, self.exvalue = extension.split('=', 1)
        except ValueError:
            # No value, just the extype
            self.extype, self.exvalue = extension, None
        else:
            self.exvalue = url_unquote(self.exvalue.strip(), encoding='utf-8')
        self.extype = self.extype.strip()

    def unparse(self) -> str:
        """
        generate string representation of single LDAP URL extension
        """
        if self.exvalue is None:
            return '%s%s' % ('!'*bool(self.critical), self.extype)
        return '%s%s=%s' % (
            '!'*bool(self.critical),
            self.extype,
            escape_str(self.exvalue or ''),
        )

    def __str__(self) -> str:
        return self.unparse()

    def __repr__(self) -> str:
        return '<%s.%s instance at %s: %s>' % (
            self.__class__.__module__,
            self.__class__.__name__,
            hex(id(self)),
            self.__dict__
        )

    def __eq__(self, other) -> bool:
        return \
            (self.critical == other.critical) and \
            (self.extype == other.extype) and \
            (self.exvalue == other.exvalue)

    def __ne__(self, other) -> bool:
        return not self.__eq__(other)


class LDAPUrlExtensions(UserDict):
    """
    Models a collection of LDAP URL extensions as
    dictionary type
    """

    def __init__(self, default: Optional[LDAPUrlExtension] = None):
        UserDict.__init__(self)
        for key, val in (default or {}).items():
            self[key] = val

    def __setitem__(self, name: str, value: LDAPUrlExtension):
        """
        value
            Either LDAPUrlExtension instance, (critical,exvalue)
            or string'ed exvalue
        """
        assert isinstance(value, LDAPUrlExtension)
        assert name == value.extype
        self.data[name] = value

    def values(self):
        return [
            self[key]
            for key in self.keys()
        ]

    def __str__(self) -> str:
        return ','.join(map(str, self.values()))

    def __repr__(self) -> str:
        return '<%s.%s instance at %s: %s>' % (
            self.__class__.__module__,
            self.__class__.__name__,
            hex(id(self)),
            self.data
        )

    def __eq__(self, other) -> bool:
        assert isinstance(other, self.__class__), TypeError(
            "other has to be instance of %s" % (self.__class__,)
        )
        return self.data == other.data

    def parse(self, extListStr):
        """
        parse string into list of LDAPURLExtension instances
        """
        for extension_str in extListStr.strip().split(','):
            if extension_str:
                ext = LDAPUrlExtension(extension_str)
                self[ext.extype] = ext

    def unparse(self) -> str:
        """
        return comma-separated string representation of LDAP URL extensions
        """
        return ','.join([val.unparse() for val in self.values()])


class LDAPUrl:
    """
    Class for parsing and unparsing LDAP URLs
    as described in RFC 4516.

    Usable class attributes:
      urlscheme
          URL scheme (either ldap, ldaps or ldapi)
      hostport
          LDAP host (default '')
      dn
          String holding distinguished name (default '')
      attrs
          list of attribute types (default None)
      scope
          integer search scope for ldap-module
      filterstr
          String representation of LDAP Search Filters
          (see RFC 4515)
      extensions
          Dictionary used as extensions store
      who
          Maps automagically to bindname LDAP URL extension
      cred
          Maps automagically to X-BINDPW LDAP URL extension
    """

    attr2extype = {
        'who': 'bindname',
        'cred': 'X-BINDPW',
        'trace_level': 'trace',
        'pwd_filename': 'x-pwdfilename',
        'sasl_mech': 'x-saslmech',
        'sasl_authzid': 'x-saslauthzid',
        'sasl_realm': 'x-saslrealm',
        'start_tls': 'x-starttls',
    }
    urlscheme: Optional[str]
    hostport: Optional[str]
    dn: Optional[str]
    attrs: Optional[List[str]]
    scope: Optional[int]
    filterstr: Optional[str]
    extensions: Optional[LDAPUrlExtensions]

    def __init__(
            self,
            ldapUrl: Optional[str] = None,
            urlscheme: str = 'ldap',
            hostport: str = '',
            dn: str = '',
            attrs=None,
            scope=None,
            filterstr=None,
            extensions=None,
            who=None,
            cred=None,
        ):
        self.urlscheme = urlscheme
        self.hostport = hostport
        self.dn = dn
        self.attrs = attrs
        self.scope = scope
        self.filterstr = filterstr
        self.extensions = (extensions or LDAPUrlExtensions())
        if ldapUrl is not None:
            self._parse(ldapUrl)
        if who is not None:
            self.who = who
        if cred is not None:
            self.cred = cred

    def __eq__(self, other) -> bool:
        return \
            self.urlscheme == other.urlscheme and \
            self.hostport == other.hostport and \
            self.dn == other.dn and \
            self.attrs == other.attrs and \
            self.scope == other.scope and \
            self.filterstr == other.filterstr and \
            self.extensions == other.extensions

    def __ne__(self, other) -> bool:
        return not self.__eq__(other)

    def _parse(self, ldap_url):
        """
        parse a LDAP URL and set the class attributes
        urlscheme,host,dn,attrs,scope,filterstr,extensions
        """
        if not is_ldapurl(ldap_url):
            raise ValueError('Value %r for ldap_url does not seem to be a LDAP URL.' % (ldap_url))
        scheme, rest = ldap_url.split('://', 1)
        self.urlscheme = scheme.strip()
        if not self.urlscheme in LDAPURL_SCHEMES:
            raise ValueError('LDAP URL contains unsupported URL scheme %s.' % (self.urlscheme))
        slash_pos = rest.find('/')
        qemark_pos = rest.find('?')
        if (slash_pos == -1) and (qemark_pos == -1):
            # No / and ? found at all
            self.hostport = url_unquote(rest, encoding='utf-8')
            self.dn = ''
            return
        if slash_pos != -1 and (qemark_pos == -1 or (slash_pos < qemark_pos)):
            # Slash separates DN from hostport
            self.hostport = url_unquote(rest[:slash_pos], encoding='utf-8')
            # Eat the slash from rest
            rest = rest[slash_pos+1:]
        elif qemark_pos != 1 and (slash_pos == -1 or (slash_pos > qemark_pos)):
            # Question mark separates hostport from rest, DN is assumed to be empty
            self.hostport = url_unquote(rest[:qemark_pos], encoding='utf-8')
            # Do not eat question mark
            rest = rest[qemark_pos:]
        else:
            raise ValueError('Something completely weird happened!')
        paramlist = rest.split('?', 4)
        paramlist_len = len(paramlist)
        if paramlist_len >= 1:
            self.dn = url_unquote(paramlist[0], encoding='utf-8').strip()
        if (paramlist_len >= 2) and (paramlist[1]):
            self.attrs = url_unquote(paramlist[1].strip(), encoding='utf-8').split(',')
        if paramlist_len >= 3:
            scope = paramlist[2].strip()
            try:
                self.scope = SEARCH_SCOPE[scope]
            except KeyError:
                raise ValueError('Invalid search scope %r' % (scope,))
        if paramlist_len >= 4:
            filterstr = paramlist[3].strip()
            if not filterstr:
                self.filterstr = None
            else:
                self.filterstr = url_unquote(filterstr, encoding='utf-8')
        if paramlist_len >= 5:
            if paramlist[4]:
                self.extensions = LDAPUrlExtensions()
                self.extensions.parse(paramlist[4])
            else:
                self.extensions = None
        return

    def connect_uri(self) -> str:
        """
        Returns LDAP URL suitable to be passed to
        ldap0.ldapobject.LDAPObject()
        """
        if self.urlscheme == 'ldapi':
            # hostport part might contain slashes when ldapi:// is used
            hostport = escape_str(self.hostport)
        else:
            hostport = self.hostport
        return '%s://%s' % (self.urlscheme, hostport)

    def unparse(self) -> str:
        """
        Returns LDAP URL depending on class attributes set.
        """
        if self.attrs is None:
            attrs_str = ''
        else:
            attrs_str = ','.join(self.attrs)
        scope_str = SEARCH_SCOPE_STR[self.scope]
        if self.filterstr is None:
            filterstr = ''
        else:
            filterstr = escape_str(self.filterstr)
        dn = escape_str(self.dn)
        if self.urlscheme == 'ldapi':
            # hostport part might contain slashes when ldapi:// is used
            hostport = escape_str(self.hostport)
        else:
            hostport = self.hostport
        ldap_url = '%s://%s/%s?%s?%s?%s' % (
            self.urlscheme,
            hostport,
            dn,
            attrs_str,
            scope_str,
            filterstr,
        )
        if self.extensions:
            ldap_url = ldap_url+'?'+self.extensions.unparse()
        return ldap_url

    def htmlHREF(
            self,
            urlPrefix='',
            hrefText=None,
            hrefTarget=None
        ) -> str:
        """
        Returns a string with HTML link for this LDAP URL.

        urlPrefix
            Prefix before LDAP URL (e.g. for addressing another web-based client)
        hrefText
            link text/description
        hrefTarget
            string added as link target attribute
        """
        assert isinstance(urlPrefix, str), TypeError(
            "urlPrefix must be str, was %r" % (urlPrefix,)
        )
        if hrefText is None:
            hrefText = self.unparse()
        assert isinstance(hrefText, str), TypeError(
            "hrefText must be str, was %r" % (hrefText,)
        )
        if hrefTarget is None:
            target = ''
        else:
            assert isinstance(hrefTarget, str), TypeError(
                "hrefTarget must be str, was %r" % (hrefTarget,)
            )
            target = ' target="%s"' % hrefTarget
        return '<a%s href="%s%s">%s</a>' % (
            target,
            urlPrefix,
            self.unparse(),
            hrefText,
        )

    def __str__(self) -> str:
        return self.unparse()

    def __repr__(self) -> str:
        return '%s.%s(%r)' % (
            self.__class__.__module__,
            self.__class__.__name__,
            self.__str__(),
        )

    def __getattr__(self, name):
        if name in self.attr2extype:
            extype = self.attr2extype[name]
            if self.extensions and \
               extype in self.extensions and \
               not self.extensions[extype].exvalue is None:
                result = url_unquote(self.extensions[extype].exvalue, encoding='utf-8')
            else:
                return None
        else:
            raise AttributeError('%s has no attribute %s' % (
                self.__class__.__name__,
                name,
            ))
        return result # __getattr__()

    def __setattr__(self, name, value):
        if name in self.attr2extype:
            extype = self.attr2extype[name]
            if value is None:
                # A value of None means that extension is deleted
                delattr(self, name)
            elif value is not None:
                # Add appropriate extension
                self.extensions[extype] = LDAPUrlExtension(
                    extype=extype,
                    exvalue=url_unquote(value, encoding='utf-8'),
                )
        else:
            self.__dict__[name] = value

    def __delattr__(self, name):
        if name in self.attr2extype:
            extype = self.attr2extype[name]
            if self.extensions:
                try:
                    del self.extensions[extype]
                except KeyError:
                    pass
        else:
            del self.__dict__[name]



# -*- coding: ascii -*-
"""
ldap0.controls.syncrepl - classes for Content Synchronization Controls
(see RFC 4533)
"""

from typing import Optional, Sequence

from uuid import UUID

# Imports from pyasn1
from pyasn1.type import namedtype, namedval, univ, tag, constraint
from pyasn1.codec.ber import encoder, decoder

from . import RequestControl, ResponseControl
from ..res import IntermediateResponse, INTERMEDIATE_RESPONSE_REGISTRY


__all__ = [
    'SyncRequestControl',
    'SyncStateControl',
    'SyncDoneControl',
    'SyncInfoMessage',
    'SyncInfoNewCookie',
    'SyncInfoRefreshDelete',
    'SyncInfoRefreshPresent',
    'SyncInfoIDSet',
]


class SyncUUID(univ.OctetString):
    """
    syncUUID ::= OCTET STRING (SIZE(16))
    """
    subtypeSpec = constraint.ValueSizeConstraint(16, 16)


class SyncCookie(univ.OctetString):
    """
    syncCookie ::= OCTET STRING
    """


class SyncRequestMode(univ.Enumerated):
    """
    mode ENUMERATED {
        -- 0 unused
        refreshOnly       (1),
        -- 2 reserved
        refreshAndPersist (3)
    }
    """
    namedValues = namedval.NamedValues(
        ('refreshOnly', 1),
        ('refreshAndPersist', 3)
    )
    subtypeSpec = univ.Enumerated.subtypeSpec + constraint.SingleValueConstraint(1, 3)


class SyncRequestValue(univ.Sequence):
    """
    syncRequestValue ::= SEQUENCE {
        mode ENUMERATED {
            -- 0 unused
            refreshOnly       (1),
            -- 2 reserved
            refreshAndPersist (3)
        }
        cookie     syncCookie OPTIONAL,
        reloadHint BOOLEAN DEFAULT FALSE
    }
    """
    componentType = namedtype.NamedTypes(
        namedtype.NamedType('mode', SyncRequestMode()),
        namedtype.OptionalNamedType('cookie', SyncCookie()),
        namedtype.DefaultedNamedType('reloadHint', univ.Boolean(False))
    )


class SyncRequestControl(RequestControl):
    """
    The Sync Request Control is an LDAP Control [RFC4511] where the
    controlType is the object identifier 1.3.6.1.4.1.4203.1.9.1.1 and the
    controlValue, an OCTET STRING, contains a BER-encoded
    syncRequestValue.  The criticality field is either TRUE or FALSE.
    [...]
    The Sync Request Control is only applicable to the SearchRequest
    Message.
    """
    controlType = '1.3.6.1.4.1.4203.1.9.1.1'

    def __init__(self, criticality=1, cookie=None, mode='refreshOnly', reloadHint=False):
        self.criticality = criticality
        self.cookie = cookie
        self.mode = mode
        self.reloadHint = reloadHint

    def encode(self):
        rcv = SyncRequestValue()
        rcv.setComponentByName('mode', SyncRequestMode(self.mode))
        if self.cookie is not None:
            rcv.setComponentByName('cookie', SyncCookie(self.cookie))
        if self.reloadHint is not None:
            rcv.setComponentByName('reloadHint', univ.Boolean(self.reloadHint))
        return encoder.encode(rcv)


class SyncStateOp(univ.Enumerated):
    """
    state ENUMERATED {
        present (0),
        add (1),
        modify (2),
        delete (3)
    }
    """
    namedValues = namedval.NamedValues(
        ('present', 0),
        ('add', 1),
        ('modify', 2),
        ('delete', 3)
    )
    subtypeSpec = univ.Enumerated.subtypeSpec + constraint.SingleValueConstraint(0, 1, 2, 3)


class SyncStateValue(univ.Sequence):
    """
    syncStateValue ::= SEQUENCE {
        state ENUMERATED {
            present (0),
            add (1),
            modify (2),
            delete (3)
        },
        entryUUID syncUUID,
        cookie    syncCookie OPTIONAL
    }
    """
    componentType = namedtype.NamedTypes(
        namedtype.NamedType('state', SyncStateOp()),
        namedtype.NamedType('entryUUID', SyncUUID()),
        namedtype.OptionalNamedType('cookie', SyncCookie())
    )


class SyncStateControl(ResponseControl):
    """
    The Sync State Control is an LDAP Control [RFC4511] where the
    controlType is the object identifier 1.3.6.1.4.1.4203.1.9.1.2 and the
    controlValue, an OCTET STRING, contains a BER-encoded SyncStateValue.
    The criticality is FALSE.
    [...]
    The Sync State Control is only applicable to SearchResultEntry and
    SearchResultReference Messages.
    """
    controlType = '1.3.6.1.4.1.4203.1.9.1.2'
    opnames = ('present', 'add', 'modify', 'delete')

    def decode(self, encodedControlValue: bytes):
        d, _ = decoder.decode(encodedControlValue, asn1Spec=SyncStateValue())
        self.state = self.__class__.opnames[int(d['state'])]
        uuid = UUID(bytes=bytes(d['entryUUID']))
        self.entryUUID = str(uuid)
        cookie = d['cookie']
        if cookie.hasValue():
            self.cookie = bytes(cookie)
        else:
            self.cookie = None

    def __repr__(self):
        return '{}(state={}, entryUUID={!r}{})'.format(
            self.__class__.__name__,
            self.state,
            self.entryUUID,
            '' if self.cookie is None else ', cookie={!r}'.format(self.cookie),
        )


class SyncDoneValue(univ.Sequence):
    """
    syncDoneValue ::= SEQUENCE {
        cookie          syncCookie OPTIONAL,
        refreshDeletes  BOOLEAN DEFAULT FALSE
    }
    """
    componentType = namedtype.NamedTypes(
        namedtype.OptionalNamedType('cookie', SyncCookie()),
        namedtype.DefaultedNamedType('refreshDeletes', univ.Boolean(False))
    )


class SyncDoneControl(ResponseControl):
    """
    The Sync Done Control is an LDAP Control [RFC4511] where the
    controlType is the object identifier 1.3.6.1.4.1.4203.1.9.1.3 and the
    controlValue contains a BER-encoded syncDoneValue.  The criticality
    is FALSE (and hence absent).
    [...]
    The Sync Done Control is only applicable to the SearchResultDone
    Message.
    """
    controlType = '1.3.6.1.4.1.4203.1.9.1.3'

    def decode(self, encodedControlValue: bytes):
        d, _ = decoder.decode(encodedControlValue, asn1Spec=SyncDoneValue())
        cookie = d['cookie']
        if cookie.hasValue():
            self.cookie = bytes(cookie)
        else:
            self.cookie = None
        self.refreshDeletes = bool(d['refreshDeletes'])

    def __repr__(self):
        return '{}({}refreshDeletes={})'.format(
            self.__class__.__name__,
            '' if self.cookie is None else 'cookie={!r}, '.format(self.cookie),
            self.refreshDeletes,
        )


class RefreshDelete(univ.Sequence):
    """
    refreshDelete  [1] SEQUENCE {
        cookie         syncCookie OPTIONAL,
        refreshDone    BOOLEAN DEFAULT TRUE
    }
    """
    componentType = namedtype.NamedTypes(
        namedtype.OptionalNamedType('cookie', SyncCookie()),
        namedtype.DefaultedNamedType('refreshDone', univ.Boolean(True))
    )


class RefreshPresent(univ.Sequence):
    """
    refreshPresent [2] SEQUENCE {
        cookie         syncCookie OPTIONAL,
        refreshDone    BOOLEAN DEFAULT TRUE
    }
    """
    componentType = namedtype.NamedTypes(
        namedtype.OptionalNamedType('cookie', SyncCookie()),
        namedtype.DefaultedNamedType('refreshDone', univ.Boolean(True))
    )


class SyncUUIDs(univ.SetOf):
    """
    syncUUIDs      SET OF syncUUID
    """
    componentType = SyncUUID()


class SyncIdSet(univ.Sequence):
    """
    syncIdSet      [3] SEQUENCE {
        cookie         syncCookie OPTIONAL,
        refreshDeletes BOOLEAN DEFAULT FALSE,
        syncUUIDs      SET OF syncUUID
    }
    """
    componentType = namedtype.NamedTypes(
        namedtype.OptionalNamedType('cookie', SyncCookie()),
        namedtype.DefaultedNamedType('refreshDeletes', univ.Boolean(False)),
        namedtype.NamedType('syncUUIDs', SyncUUIDs())
    )


class SyncInfoValue(univ.Choice):
    """
    syncInfoValue ::= CHOICE {
        newcookie      [0] syncCookie,
        refreshDelete  [1] SEQUENCE {
            cookie         syncCookie OPTIONAL,
            refreshDone    BOOLEAN DEFAULT TRUE
        },
        refreshPresent [2] SEQUENCE {
            cookie         syncCookie OPTIONAL,
            refreshDone    BOOLEAN DEFAULT TRUE
        },
        syncIdSet      [3] SEQUENCE {
            cookie         syncCookie OPTIONAL,
            refreshDeletes BOOLEAN DEFAULT FALSE,
            syncUUIDs      SET OF syncUUID
        }
    }
    """
    componentType = namedtype.NamedTypes(
        namedtype.NamedType(
            'newcookie',
            SyncCookie().subtype(
                implicitTag=tag.Tag(tag.tagClassContext, tag.tagFormatConstructed, 0)
            )
        ),
        namedtype.NamedType(
            'refreshDelete',
            RefreshDelete().subtype(
                implicitTag=tag.Tag(tag.tagClassContext, tag.tagFormatConstructed, 1)
            )
        ),
        namedtype.NamedType(
            'refreshPresent',
            RefreshPresent().subtype(
                implicitTag=tag.Tag(tag.tagClassContext, tag.tagFormatConstructed, 2)
            )
        ),
        namedtype.NamedType(
            'syncIdSet',
            SyncIdSet().subtype(
                implicitTag=tag.Tag(tag.tagClassContext, tag.tagFormatConstructed, 3)
            )
        )
    )


class SyncInfoMessage(IntermediateResponse):
    """
    The Sync Info Message is an LDAP Intermediate Response Message
    [RFC4511] where responseName is the object identifier
    1.3.6.1.4.1.4203.1.9.1.4 and responseValue contains a BER-encoded
    syncInfoValue.  The criticality is FALSE (and hence absent).
    """
    responseName = '1.3.6.1.4.1.4203.1.9.1.4'

    def __new__(cls, responseName, encodedResponseValue, ctrls):
        if cls is not __class__:
            return super().__new__(cls)
        syncinfo, _ = decoder.decode(encodedResponseValue, asn1Spec=SyncInfoValue())
        choice = syncinfo.getName()
        if choice == 'newcookie':
            child = SyncInfoNewCookie
        elif choice == 'refreshDelete':
            child = SyncInfoRefreshDelete
        elif choice == 'refreshPresent':
            child = SyncInfoRefreshPresent
        elif choice == 'syncIdSet':
            child = SyncInfoIDSet
        else:
            raise ValueError
        return child.__new__(child, responseName, encodedResponseValue, ctrls)

    def decode(self, value: bytes):
        self.syncinfo, _ = decoder.decode(
            value,
            asn1Spec=SyncInfoValue(),
        )

class SyncInfoNewCookie(SyncInfoMessage):
    cookie: bytes

    def decode(self, value: bytes):
        super().decode(value)
        self.cookie = bytes(self.syncinfo.getComponent())

    def __repr__(self):
        return '{}(cookie={!r})'.format(
            self.__class__.__name__,
            self.cookie,
        )


class SyncInfoRefreshDelete(SyncInfoMessage):
    cookie: Optional[bytes]
    refreshDone: bool

    def decode(self, value: bytes):
        super().decode(value)
        component = self.syncinfo.getComponent()
        self.cookie = None
        cookie = component['cookie']
        if cookie.isValue:
            self.cookie = bytes(cookie)
        self.refreshDone = bool(component['refreshDone'])

    def __repr__(self):
        return '{}(cookie={!r}, refreshDone={!r})'.format(
            self.__class__.__name__,
            self.cookie,
            self.refreshDone,
        )


class SyncInfoRefreshPresent(SyncInfoMessage):
    cookie: Optional[bytes]
    refreshDone: bool

    def decode(self, value: bytes):
        super().decode(value)
        component = self.syncinfo.getComponent()
        self.cookie = None
        cookie = component['cookie']
        if cookie.isValue:
            self.cookie = bytes(cookie)
        self.refreshDone = bool(component['refreshDone'])

    def __repr__(self):
        return '{}(cookie={!r}, refreshDone={!r})'.format(
            self.__class__.__name__,
            self.cookie,
            self.refreshDone,
        )


class SyncInfoIDSet(SyncInfoMessage):
    cookie: Optional[bytes]
    refreshDeletes: bool
    syncUUIDs: Sequence[str]

    def decode(self, value: bytes):
        super().decode(value)
        component = self.syncinfo.getComponent()
        self.cookie = None
        cookie = component['cookie']
        if cookie.isValue:
            self.cookie = bytes(cookie)
        self.refreshDeletes = bool(component['refreshDeletes'])

        uuids = []
        for syncuuid in component['syncUUIDs']:
            uuid = UUID(bytes=bytes(syncuuid))
            uuids.append(str(uuid))
        self.syncUUIDs = uuids

    def __repr__(self):
        return '{}(cookie={!r}, refreshDeletes={!r}, syncUUIDs={!r})'.format(
            self.__class__.__name__,
            self.cookie,
            self.refreshDeletes,
            self.syncUUIDs,
        )


INTERMEDIATE_RESPONSE_REGISTRY.register(SyncInfoMessage)



# -*- coding: utf-8 -*-
"""
Automatic tests for module ldap0.ldapurl
"""

import unittest
from urllib.request import quote as url_quote

import ldap0
import ldap0.ldapurl
from ldap0.ldapurl import LDAPUrl


class TestIsLDAPUrl(unittest.TestCase):

    is_ldapurl_tests = {
        # Examples from RFC2255
        'ldap:///o=University%20of%20Michigan,c=US':1,
        'ldap://ldap.itd.umich.edu/o=University%20of%20Michigan,c=US':1,
        'ldap://ldap.itd.umich.edu/o=University%20of%20Michigan,':1,
        'ldap://host.com:6666/o=University%20of%20Michigan,':1,
        'ldap://ldap.itd.umich.edu/c=GB?objectClass?one':1,
        'ldap://ldap.question.com/o=Question%3f,c=US?mail':1,
        'ldap://ldap.netscape.com/o=Babsco,c=US??(int=%5c00%5c00%5c00%5c04)':1,
        'ldap:///??sub??bindname=cn=Manager%2co=Foo':1,
        'ldap:///??sub??!bindname=cn=Manager%2co=Foo':1,
        # More examples from various sources
        'ldap://ldap.nameflow.net:1389/c%3dDE':1,
        'ldap://root.openldap.org/dc=openldap,dc=org':1,
        'ldap://root.openldap.org/dc=openldap,dc=org':1,
        'ldap://x500.mh.se/o=Mitthogskolan,c=se????1.2.752.58.10.2=T.61':1,
        'ldp://root.openldap.org/dc=openldap,dc=org':0,
        'ldap://localhost:1389/ou%3DUnstructured%20testing%20tree%2Cdc%3Dexample%2Cdc%3Dcom??one':1,
        'ldaps://ldap.example.com/c%3dDE':1,
        'ldapi:///dc=example,dc=com????x-saslmech=EXTERNAL':1,
    }

    def test_is_ldapurl(self):
        for ldap_url, expected in self.is_ldapurl_tests.items():
            result = ldap0.ldapurl.is_ldapurl(ldap_url)
            self.assertEqual(
                result, expected,
                'is_ldapurl("%s") returns %d instead of %d.' % (
                    ldap_url, result, expected,
                )
            )


class TestParseLDAPUrl(unittest.TestCase):

    parse_ldap_url_tests = [
        (
            'ldap://root.openldap.org/dc=openldap,dc=org',
            LDAPUrl(
                hostport='root.openldap.org',
                dn='dc=openldap,dc=org'
            )
        ),
        (
            'ldap://root.openldap.org/dc%3dboolean%2cdc%3dnet???%28objectClass%3d%2a%29',
            LDAPUrl(
                hostport='root.openldap.org',
                dn='dc=boolean,dc=net',
                filterstr='(objectClass=*)'
            )
        ),
        (
            'ldap://root.openldap.org/dc=openldap,dc=org??sub?',
            LDAPUrl(
                hostport='root.openldap.org',
                dn='dc=openldap,dc=org',
                scope=ldap0.SCOPE_SUBTREE
            )
        ),
        (
            'ldap://root.openldap.org/dc=openldap,dc=org??one?',
            LDAPUrl(
                hostport='root.openldap.org',
                dn='dc=openldap,dc=org',
                scope=ldap0.SCOPE_ONELEVEL
            )
        ),
        (
            'ldap://root.openldap.org/dc=openldap,dc=org??base?',
            LDAPUrl(
                hostport='root.openldap.org',
                dn='dc=openldap,dc=org',
                scope=ldap0.SCOPE_BASE
            )
        ),
        (
            'ldap://x500.mh.se/o=Mitthogskolan,c=se????1.2.752.58.10.2=T.61',
            LDAPUrl(
                hostport='x500.mh.se',
                dn='o=Mitthogskolan,c=se',
                extensions=ldap0.ldapurl.LDAPUrlExtensions({
                    '1.2.752.58.10.2':ldap0.ldapurl.LDAPUrlExtension(
                    critical=0,extype='1.2.752.58.10.2',exvalue='T.61'
                    )
                })
            )
        ),
        (
            'ldap://localhost:12345/dc=example,dc=com????!bindname=cn=Michael%2Cdc=example%2Cdc=com,!X-BINDPW=secretpassword',
            LDAPUrl(
                hostport='localhost:12345',
                dn='dc=example,dc=com',
                extensions=ldap0.ldapurl.LDAPUrlExtensions({
                    'bindname':ldap0.ldapurl.LDAPUrlExtension(
                    critical=1,extype='bindname',exvalue='cn=Michael,dc=example,dc=com'
                    ),
                    'X-BINDPW':ldap0.ldapurl.LDAPUrlExtension(
                    critical=1,extype='X-BINDPW',exvalue='secretpassword'
                    ),
                }),
            )
        ),
        (
            'ldap://localhost:54321/dc=example,dc=com????bindname=cn=Michael%2Cdc=example%2Cdc=com,X-BINDPW=secretpassword',
            LDAPUrl(
                hostport='localhost:54321',
                dn='dc=example,dc=com',
                who='cn=Michael,dc=example,dc=com',
                cred='secretpassword'
            )
        ),
        (
            'ldap://localhost:54321/cn=Michael Ströder+mail=michael@example.com,o=Test????bindname=cn=Michael%20Ströder%2bmail=michael@example.com%2Co=Test,X-BINDPW=secretpassword',
            LDAPUrl(
                hostport='localhost:54321',
                dn='cn=Michael Ströder+mail=michael@example.com,o=Test',
                who='cn=Michael Ströder+mail=michael@example.com,o=Test',
                cred='secretpassword'
            )
        ),
        (
            'ldaps://localhost:12345/dc=example,dc=com',
            LDAPUrl(
                urlscheme='ldaps',
                hostport='localhost:12345',
                dn='dc=example,dc=com',
            ),
        ),
        (
            'ldapi://%2ftmp%2fopenldap2-1389/dc=example,dc=com',
            LDAPUrl(
                urlscheme='ldapi',
                hostport='/tmp/openldap2-1389',
                dn='dc=example,dc=com',
            ),
        ),
    ]

    def test_ldapurl(self):
        for ldap_url_str,test_ldap_url_obj in self.parse_ldap_url_tests:
            ldap_url_obj = LDAPUrl(ldapUrl=ldap_url_str)
            self.assertEqual(
                ldap_url_obj, test_ldap_url_obj,
                'Attributes of LDAPUrl(%r) are:\n%r\ninstead of:\n%r' % (
                    ldap_url_str,
                    ldap_url_obj,
                    test_ldap_url_obj,
                )
            )
            unparsed_ldap_url_str = test_ldap_url_obj.unparse()
            unparsed_ldap_url_obj = LDAPUrl(ldapUrl=unparsed_ldap_url_str)
            self.assertEqual(
                unparsed_ldap_url_obj, test_ldap_url_obj,
                'Attributes of LDAPUrl(%r) are:\n%r\ninstead of:\n%r' % (
                    unparsed_ldap_url_str,
                    unparsed_ldap_url_obj,
                    test_ldap_url_obj,
                )
            )


class TestLDAPUrl(unittest.TestCase):

    def test_combo(self):
        u = LDAPUrl(
            'ldap://127.0.0.1:1234/'
            'dc=example,dc=com'
            '?attr1,attr2,attr3'
            '?sub'
            '?%28objectClass%3D%2A%29'
            '?'
            'bindname=cn%3Dfred%2Cc%3Dau,'
            'X-BINDPW=%3F%3F%3F,'
            'trace=8,'
            'x-saslmech=DIGEST-MD5,'
            'x-saslrealm=example.com,'
            'x-saslauthzid=u:anna,'
            'x-starttls=2,'
        )
        self.assertEqual(u.urlscheme, "ldap")
        self.assertEqual(u.hostport, "127.0.0.1:1234")
        self.assertEqual(u.dn, "dc=example,dc=com")
        self.assertEqual(u.attrs, ["attr1","attr2","attr3"])
        self.assertEqual(u.scope, ldap0.SCOPE_SUBTREE)
        self.assertEqual(u.filterstr, "(objectClass=*)")
        self.assertEqual(len(u.extensions), 7)
        self.assertEqual(u.who, "cn=fred,c=au")
        self.assertEqual(u.cred, "???")
        self.assertEqual(u.trace_level, "8")
        self.assertEqual(u.sasl_mech, 'DIGEST-MD5')
        self.assertEqual(u.sasl_realm, 'example.com')
        self.assertEqual(u.sasl_authzid, 'u:anna')
        self.assertEqual(u.start_tls, '2')
        with self.assertRaises(AttributeError):
            u.foo

    def test_parse_default_hostport(self):
        u = LDAPUrl("ldap://")
        self.assertEqual(u.urlscheme, "ldap")
        self.assertEqual(u.hostport, "")

    def test_parse_empty_dn(self):
        u = LDAPUrl("ldap://")
        self.assertEqual(u.dn, "")
        u = LDAPUrl("ldap:///")
        self.assertEqual(u.dn, "")
        u = LDAPUrl("ldap:///?")
        self.assertEqual(u.dn, "")

    def test_parse_default_attrs(self):
        u = LDAPUrl("ldap://")
        self.assertIsNone(u.attrs)

    def test_parse_default_scope(self):
        u = LDAPUrl("ldap://")
        self.assertIsNone(u.scope)     # RFC4516 s3

    def test_parse_default_filter(self):
        u = LDAPUrl("ldap://")
        self.assertIsNone(u.filterstr) # RFC4516 s3

    def test_parse_default_extensions(self):
        u = LDAPUrl("ldap://")
        self.assertEqual(len(u.extensions), 0)

    def test_parse_schemes(self):
        u = LDAPUrl("ldap://")
        self.assertEqual(u.urlscheme, "ldap")
        u = LDAPUrl("ldapi://")
        self.assertEqual(u.urlscheme, "ldapi")
        u = LDAPUrl("ldaps://")
        self.assertEqual(u.urlscheme, "ldaps")

    def test_parse_hostport(self):
        u = LDAPUrl("ldap://a")
        self.assertEqual(u.hostport, "a")
        u = LDAPUrl("ldap://a.b")
        self.assertEqual(u.hostport, "a.b")
        u = LDAPUrl("ldap://a.")
        self.assertEqual(u.hostport, "a.")
        u = LDAPUrl("ldap://%61%62:%32/")
        self.assertEqual(u.hostport, "ab:2")
        u = LDAPUrl("ldap://[::1]/")
        self.assertEqual(u.hostport, "[::1]")
        u = LDAPUrl("ldap://[::1]")
        self.assertEqual(u.hostport, "[::1]")
        u = LDAPUrl("ldap://[::1]:123/")
        self.assertEqual(u.hostport, "[::1]:123")
        u = LDAPUrl("ldap://[::1]:123")
        self.assertEqual(u.hostport, "[::1]:123")

    def test_parse_dn(self):
        u = LDAPUrl("ldap:///")
        self.assertEqual(u.dn, "")
        u = LDAPUrl("ldap:///dn=foo")
        self.assertEqual(u.dn, "dn=foo")
        u = LDAPUrl("ldap:///dn=foo%2cdc=bar")
        self.assertEqual(u.dn, "dn=foo,dc=bar")
        u = LDAPUrl("ldap:///dn=foo%20bar")
        self.assertEqual(u.dn, "dn=foo bar")
        u = LDAPUrl("ldap:///dn=foo%2fbar")
        self.assertEqual(u.dn, "dn=foo/bar")
        u = LDAPUrl("ldap:///dn=foo%2fbar?")
        self.assertEqual(u.dn, "dn=foo/bar")
        u = LDAPUrl("ldap:///dn=foo%3f?")
        self.assertEqual(u.dn, "dn=foo?")
        u = LDAPUrl("ldap:///dn=foo%3f")
        self.assertEqual(u.dn, "dn=foo?")
        u = LDAPUrl("ldap:///dn=str%c3%b6der.com")
        self.assertEqual(u.dn, "dn=ströder.com")

    def test_parse_attrs(self):
        u = LDAPUrl("ldap:///?")
        self.assertEqual(u.attrs, None)
        u = LDAPUrl("ldap:///??")
        self.assertEqual(u.attrs, None)
        u = LDAPUrl("ldap:///?*?")
        self.assertEqual(u.attrs, ['*'])
        u = LDAPUrl("ldap:///?*,*?")
        self.assertEqual(u.attrs, ['*','*'])
        u = LDAPUrl("ldap:///?a")
        self.assertEqual(u.attrs, ['a'])
        u = LDAPUrl("ldap:///?%61")
        self.assertEqual(u.attrs, ['a'])
        u = LDAPUrl("ldap:///?a,b")
        self.assertEqual(u.attrs, ['a','b'])
        u = LDAPUrl("ldap:///?a%3fb")
        self.assertEqual(u.attrs, ['a?b'])

    def test_parse_scope_default(self):
        u = LDAPUrl("ldap:///??")
        self.assertIsNone(u.scope) # on opposite to RFC4516 s3 for referral chasing
        u = LDAPUrl("ldap:///???")
        self.assertIsNone(u.scope) # on opposite to RFC4516 s3 for referral chasing

    def test_parse_scope(self):
        u = LDAPUrl("ldap:///??sub")
        self.assertEqual(u.scope, ldap0.SCOPE_SUBTREE)
        u = LDAPUrl("ldap:///??sub?")
        self.assertEqual(u.scope, ldap0.SCOPE_SUBTREE)
        u = LDAPUrl("ldap:///??base")
        self.assertEqual(u.scope, ldap0.SCOPE_BASE)
        u = LDAPUrl("ldap:///??base?")
        self.assertEqual(u.scope, ldap0.SCOPE_BASE)
        u = LDAPUrl("ldap:///??one")
        self.assertEqual(u.scope, ldap0.SCOPE_ONELEVEL)
        u = LDAPUrl("ldap:///??one?")
        self.assertEqual(u.scope, ldap0.SCOPE_ONELEVEL)
        u = LDAPUrl("ldap:///??subordinates")
        self.assertEqual(u.scope, ldap0.SCOPE_SUBORDINATE)
        u = LDAPUrl("ldap:///??subordinates?")
        self.assertEqual(u.scope, ldap0.SCOPE_SUBORDINATE)

    def test_parse_filter(self):
        u = LDAPUrl("ldap:///???(cn=Bob)")
        self.assertEqual(u.filterstr, "(cn=Bob)")
        u = LDAPUrl("ldap:///???(cn=Bob)?")
        self.assertEqual(u.filterstr, "(cn=Bob)")
        u = LDAPUrl("ldap:///???(cn=Bob%20Smith)?")
        self.assertEqual(u.filterstr, "(cn=Bob Smith)")
        u = LDAPUrl("ldap:///???(cn=Bob/Smith)?")
        self.assertEqual(u.filterstr, "(cn=Bob/Smith)")
        u = LDAPUrl("ldap:///???(cn=Bob:Smith)?")
        self.assertEqual(u.filterstr, "(cn=Bob:Smith)")
        u = LDAPUrl("ldap:///???&(cn=Bob)(objectClass=user)?")
        self.assertEqual(u.filterstr, "&(cn=Bob)(objectClass=user)")
        u = LDAPUrl("ldap:///???|(cn=Bob)(objectClass=user)?")
        self.assertEqual(u.filterstr, "|(cn=Bob)(objectClass=user)")
        u = LDAPUrl("ldap:///???(cn=Q%3f)?")
        self.assertEqual(u.filterstr, "(cn=Q?)")
        u = LDAPUrl("ldap:///???(cn=Q%3f)")
        self.assertEqual(u.filterstr, "(cn=Q?)")
        u = LDAPUrl("ldap:///???(sn=Str%c3%b6der)") # (possibly bad?)
        self.assertEqual(u.filterstr, "(sn=Ströder)")
        u = LDAPUrl("ldap:///???(sn=Str\\c3\\b6der)")
        self.assertEqual(u.filterstr, "(sn=Str\\c3\\b6der)") # (recommended)
        u = LDAPUrl("ldap:///???(cn=*\\2a*)")
        self.assertEqual(u.filterstr, "(cn=*\\2a*)")
        u = LDAPUrl("ldap:///???(cn=*%5c2a*)")
        self.assertEqual(u.filterstr, "(cn=*\\2a*)")

    def test_parse_extensions(self):
        u = LDAPUrl("ldap:///????")
        self.assertIsNone(u.extensions)
        self.assertIsNone(u.who)
        u = LDAPUrl("ldap:///????bindname=cn=root")
        self.assertEqual(len(u.extensions), 1)
        self.assertEqual(u.who, "cn=root")
        u = LDAPUrl("ldap:///????!bindname=cn=root")
        self.assertEqual(len(u.extensions), 1)
        self.assertEqual(u.who, "cn=root")
        u = LDAPUrl("ldap:///????bindname=%3f,X-BINDPW=%2c")
        self.assertEqual(len(u.extensions), 2)
        self.assertEqual(u.who, "?")
        self.assertEqual(u.cred, ",")

    def test_parse_extensions_nulls(self):
        u = LDAPUrl("ldap:///????bindname=%00name")
        self.assertEqual(u.who, "\0name")

    def test_parse_extensions_5questions(self):
        u = LDAPUrl("ldap:///????bindname=?")
        self.assertEqual(len(u.extensions), 1)
        self.assertEqual(u.who, "?")

    def test_parse_extensions_novalue(self):
        u = LDAPUrl("ldap:///????bindname")
        self.assertEqual(len(u.extensions), 1)
        self.assertIsNone(u.who)

    def test_bad_urls1(self):
        failed_urls = []
        for bad in (
                '',
                'ldap:',
                'ldap:/',
                ':///',
                '://',
                '///',
                '//',
                '/',
                'LDAP://',
                'invalid://',
                'ldap:///??invalid_scope',
                r'ldap:///??%00',       # RFC4516 2.1
        ):
            try:
                LDAPUrl(bad)
            except ValueError:
                pass
            else:
                failed_urls.append(bad)
        if failed_urls:
            self.fail("These LDAP URLs should have raised ValueError: %r" % failed_urls)

    @unittest.expectedFailure
    def test_bad_urls2(self):
        failed_urls = []
        for bad in (
                #XXX-- the following should raise exceptions!
                'ldap:///?????',       # extension can't start with '?'
                'ldap://:389/',         # [host [COLON port]]
                'ldap://a:/',           # [host [COLON port]]
                r'ldap://%%%/',          # invalid URL encoding
                'ldap:///?,',           # attrdesc *(COMMA attrdesc)
                'ldap:///?a,',          # attrdesc *(COMMA attrdesc)
                'ldap:///?,a',          # attrdesc *(COMMA attrdesc)
                'ldap:///?a,,b',        # attrdesc *(COMMA attrdesc)
                r'ldap://%00/',         # RFC4516 2.1
                r'ldap:///%00',         # RFC4516 2.1
                r'ldap:///?%00',        # RFC4516 2.1
                'ldap:///????0=0',      # extype must start with Alpha
                'ldap:///????a_b=0',    # extype contains only [-a-zA-Z0-9]
                'ldap:///????!!a=0',    # only one exclamation allowed
            ):
            try:
                LDAPUrl(bad)
            except ValueError:
                pass
            else:
                failed_urls.append(bad)
        if failed_urls:
            self.fail("These LDAP URLs should have raised ValueError: %r" % failed_urls)

    def test_connect_uri(self):
        for url, uri in (
            ('ldap:///o=University%20of%20Michigan,c=US', 'ldap://'),
            ('ldap://ldap.itd.umich.edu/o=University%20of%20Michigan,c=US', 'ldap://ldap.itd.umich.edu'),
            ('ldap://host.com:6666/o=University%20of%20Michigan,', 'ldap://host.com:6666'),
            ('ldap://ldap.nameflow.net:1389/c%3dDE', 'ldap://ldap.nameflow.net:1389'),
            ('ldap://root.openldap.org/dc=openldap,dc=org', 'ldap://root.openldap.org'),
            ('ldap://x500.mh.se/o=Mitthogskolan,c=se????1.2.752.58.10.2=T.61', 'ldap://x500.mh.se'),
            ('ldap://localhost:1389/ou%3DUnstructured%20testing%20tree??one', 'ldap://localhost:1389'),
            ('ldapi:///dc=example,dc=com????x-saslmech=EXTERNAL', 'ldapi://'),
            ('ldapi://%2Fopt%2Fae-dir%2Frun%2Fslapd%2Fldapi/ou=ae-dir????bindname=aead', 'ldapi://%2Fopt%2Fae-dir%2Frun%2Fslapd%2Fldapi'),
        ):
            self.assertEqual(LDAPUrl(url).connect_uri(), uri)


if __name__ == '__main__':
    unittest.main()



# -*- coding: ascii -*-
"""
ldap0.typing - specific type hints
"""

from typing import Dict, List, Sequence, Tuple

from .controls import RequestControl


BytesList = List[bytes]
StrList = List[str]

AttrList = Sequence[str]

EntryBytes = Dict[bytes, BytesList]
EntryMixed = Dict[str, BytesList]
EntryStr = Dict[str, StrList]

ModList = List[Tuple[int, bytes, BytesList]]

RequestControls = Sequence[RequestControl]



# -*- coding: utf-8 -*-
"""
Automatic tests for module ldap0.schema.parse
"""

import unittest

from ldap0.schema.parse import split_tokens

# basic test cases
TESTCASES_BASIC = (
    (" BLUBBER DI BLUBB ", ["BLUBBER", "DI", "BLUBB"]),
    ("BLUBBER DI BLUBB", ["BLUBBER", "DI", "BLUBB"]),
    ("BL-UB-BER DI BL-UBB", ["BL-UB-BER", "DI", "BL-UBB"]),
    ("BLUBBER  DI   BLUBB  ", ["BLUBBER", "DI", "BLUBB"]),
    ("BLUBBER  DI  'BLUBB'   ", ["BLUBBER", "DI", "BLUBB"]),
    ("BLUBBER ( DI ) 'BLUBB'   ", ["BLUBBER", "(", "DI", ")", "BLUBB"]),
    ("BLUBBER(DI)", ["BLUBBER", "(", "DI", ")"]),
    ("BLUBBER ( DI)", ["BLUBBER", "(", "DI", ")"]),
    ("BLUBBER ''", ["BLUBBER", ""]),
    ("( BLUBBER (DI 'BLUBB'))", ["(", "BLUBBER", "(", "DI", "BLUBB", ")", ")"]),
    ("BLUBB (DA$BLAH)", ['BLUBB', "(", "DA", "BLAH", ")"]),
    ("BLUBB ( DA $  BLAH )", ['BLUBB', "(", "DA", "BLAH", ")"]),
    ("BLUBB (DA$ BLAH)", ['BLUBB', "(", "DA", "BLAH", ")"]),
    ("BLUBB (DA $BLAH)", ['BLUBB', "(", "DA", "BLAH", ")"]),
    ("BLUBB 'DA$BLAH'", ['BLUBB', "DA$BLAH"]),
    ("BLUBB DI 'BLU B B ER' DA 'BLAH' ", ['BLUBB', 'DI', 'BLU B B ER', 'DA', 'BLAH']),
    ("BLUBB DI 'BLU B B ER' DA 'BLAH' LABER", ['BLUBB', 'DI', 'BLU B B ER', 'DA', 'BLAH', 'LABER']),
    ("BLUBB\t'DA\tBLUB'", ['BLUBB', "DA\tBLUB"]),
)

# UTF-8 raw strings
TESTCASES_UTF8 = (
    (" BL\xc3\x9cBBER D\xc3\x84 BL\xc3\x9cBB ", ["BL\xc3\x9cBBER", "D\xc3\x84", "BL\xc3\x9cBB"]),
    ("BL\xc3\x9cBBER D\xc3\x84 BL\xc3\x9cBB", ["BL\xc3\x9cBBER", "D\xc3\x84", "BL\xc3\x9cBB"]),
    ("BL\xc3\x9cBBER  D\xc3\x84   BL\xc3\x9cBB  ", ["BL\xc3\x9cBBER", "D\xc3\x84", "BL\xc3\x9cBB"]),
)

# broken schema of Oracle Internet Directory
TESTCASES_BROKEN_OID = (
    ("BLUBB DI 'BLU B B ER'MUST 'BLAH' ", ['BLUBB', 'DI', 'BLU B B ER', 'MUST', 'BLAH']),
    ("BLUBBER DI 'BLU'BB ER' DA 'BLAH' ", ["BLUBBER", "DI", "BLU'BB ER", "DA", "BLAH"]),
)

# for quoted single quotes inside string values
TESTCASES_ESCAPED_QUOTES = (
    ("BLUBBER '\\''", ["BLUBBER", "'"]),
    ("BLUBBER DI 'BLU\\'BB ER' DA 'BLAH' ", ["BLUBBER", "DI", "BLU'BB ER", "DA", "BLAH"]),
    ("BLUBBER DI 'BLU\\' BB ER' DA 'BLAH' ", ["BLUBBER", "DI", "BLU' BB ER", "DA", "BLAH"]),
)

# test cases which should result in ValueError raised
TESTCASES_BROKEN = (
    "( BLUB",
    "BLUB )",
    "BLUB 'DA",
    "BLUB $ DA",
#    "BLUB 'DA\\'",
#    "( BLUB )) DA (",
)

class TestSplitTokens(unittest.TestCase):
    """
    test function ldap0.schema.parse.split_tokens()
    """

    def _run_split_tokens_tests(self, test_cases):
        for test_value, test_result in test_cases:
            token_list = split_tokens(test_value)
            self.assertEqual(token_list, test_result)

    def _run_failure_tests(self, test_cases):
        should_have_failed = []
        for test_value in test_cases:
            try:
                _ = split_tokens(test_value)
            except ValueError:
                pass
            else:
                should_have_failed.append(test_value)
        if should_have_failed:
            self.fail(
                '%d value(s) should have raised ValueError: %r' % (
                    len(should_have_failed),
                    should_have_failed,
                )
            )

    def test_basic(self):
        """
        run test cases specified in constant TESTCASES_BASIC
        """
        self._run_split_tokens_tests(TESTCASES_BASIC)

    def test_utf8(self):
        """
        run test cases specified in constant TESTCASES_BASIC
        """
        self._run_split_tokens_tests(TESTCASES_UTF8)

    @unittest.expectedFailure
    def test_let_broken_oid_pass(self):
        """
        run test cases specified in constant TESTCASES_BROKEN_OID
        """
        self._run_split_tokens_tests(TESTCASES_BROKEN_OID)

    @unittest.expectedFailure
    def test_escaped_quotes(self):
        """
        run test cases specified in constant TESTCASES_ESCAPED_QUOTES
        """
        self._run_split_tokens_tests(TESTCASES_ESCAPED_QUOTES)

    def test_broken(self):
        self._run_failure_tests(TESTCASES_BROKEN)


if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*-
"""
Automatic tests for module ldap0.dn
"""

# from Python's standard lib
import sys
import unittest

import ldap0.dn
from ldap0.dn import DNObj


class TestDNFunctions(unittest.TestCase):
    """
    test functions in ldap0.dn
    """

    def test_is_dn(self):
        """
        test function is_dn()
        """
        self.assertEqual(ldap0.dn.is_dn('foobar,dc=example,dc=com'), False)
        self.assertEqual(ldap0.dn.is_dn('-cn=foobar,dc=example,dc=com'), False)
        self.assertEqual(ldap0.dn.is_dn(';cn=foobar,dc=example,dc=com'), False)
        self.assertEqual(ldap0.dn.is_dn(',cn=foobar,dc=example,dc=com'), False)
        self.assertEqual(ldap0.dn.is_dn('cn=foobar,dc=example,dc=com,'), False)
        self.assertEqual(ldap0.dn.is_dn('uid=xkcd,cn=foobar,dc=example,dc=com'), True)
        self.assertEqual(ldap0.dn.is_dn('cn=äöüÄÖÜß'), True)
        self.assertEqual(ldap0.dn.is_dn('-'), False)
        self.assertEqual(ldap0.dn.is_dn(''), True)
        self.assertEqual(ldap0.dn.is_dn('cn='), True)

    def test_escape_str(self):
        """
        test function escape_str()
        """
        self.assertEqual(ldap0.dn.escape_str(''), '')
        self.assertEqual(ldap0.dn.escape_str('foobar'), 'foobar')
        self.assertEqual(ldap0.dn.escape_str('foo,bar'), 'foo\\,bar')
        self.assertEqual(ldap0.dn.escape_str('foo=bar'), 'foo\\=bar')
        self.assertEqual(ldap0.dn.escape_str('foo#bar'), 'foo#bar')
        self.assertEqual(ldap0.dn.escape_str('#foobar'), '\\#foobar')
        self.assertEqual(ldap0.dn.escape_str('foo bar'), 'foo bar')
        self.assertEqual(ldap0.dn.escape_str(' foobar'), '\\ foobar')
        self.assertEqual(ldap0.dn.escape_str('  foobar'), '\\  foobar')
        self.assertEqual(ldap0.dn.escape_str('foobar  '), 'foobar \\ ')
        self.assertEqual(ldap0.dn.escape_str(' '), '\\ ')
        self.assertEqual(ldap0.dn.escape_str('  '), '\\ \\ ')


class TestDNObj(unittest.TestCase):
    """
    test class ldap0.dn.DNObj
    """

    def test001_simple(self):
        self.assertEqual(DNObj.from_str(''), DNObj((())))
        self.assertEqual(DNObj.from_str('').rdn(), DNObj((())))
        self.assertEqual(DNObj.from_str('').rdn_attrs(), {})
        self.assertEqual(len(DNObj.from_str('')), 0)
        self.assertEqual(len(DNObj.from_str('ou=dept1  ,  dc=example, dc=com')), 3)
        self.assertEqual(
            str(DNObj.from_str('ou=dept1  ,  dc=example, dc=com').rdn()),
            'ou=dept1'
        )
        self.assertEqual(
            DNObj.from_str('ou=dept1  ,  dc=example, dc=com').rdn_attrs(),
            {'ou': 'dept1'}
        )
        self.assertEqual(
            DNObj.from_str('ou=dept1 + mail=test@example.com ,  dc=example, dc=com').rdn_attrs(),
            {'mail': 'test@example.com', 'ou': 'dept1'}
        )
        self.assertEqual(
            str(DNObj.from_str('ou=dept1,  dc=example, dc=com').parent()),
            'dc=example,dc=com'
        )
        for dn_str in (
                '',
                'dc=example,dc=com',
                'ou=dept1,dc=example,dc=com',
                'cn=äöüÄÖÜß',
                'cn=Michael Ströder,dc=example,dc=com',
            ):
            dn_obj = DNObj.from_str(dn_str, flags=0)
            self.assertEqual(bytes(dn_obj), dn_str.encode('utf-8'))
            self.assertEqual(dn_obj.__str__(), dn_str)

    def test002_weird_whitespaces(self):
        self.assertEqual(
            DNObj.from_str('uid='),
            DNObj((
                (
                    ('uid', None),
                ),
            ))
        )
        self.assertEqual(
            DNObj.from_str('uid = test42 + uidNumber = 42, dc = example , dc = com'),
            DNObj((
                (
                    ('uid', 'test42'),
                    ('uidNumber', '42')
                ),
                (('dc', 'example'),),
                (('dc', 'com'),)
            ))
        )

    def test003_repr(self):
        if sys.version_info.major >= 3:
            correct_repr = "DNObj(((('uid', 'test42'), ('uidNumber', '42')), (('dc', 'example'),), (('dc', 'com'),)))"
        else:
            correct_repr = "DNObj(((('uid', 'test42'), ('uidNumber', '42')), (('dc', 'example'),), (('dc', 'com'),)))"
        self.assertEqual(
            repr(DNObj((
                (
                    ('uid', 'test42'),
                    ('uidNumber', '42')
                ),
                (('dc', 'example'),),
                (('dc', 'com'),)
            ))),
            correct_repr
        )

    def test004_hashable(self):
        dn = DNObj((
            (
                ('uid', 'test42'),
                ('uidNumber', '42')
            ),
            (('dc', 'example'),),
            (('dc', 'com'),)
        ))
        data = {dn: 'foo'}
        self.assertEqual(data[dn], 'foo')
        self.assertEqual(set([dn]), set([dn, dn]))

    def test005_slicing(self):
        dn = DNObj.from_str('ou=dept1  ,  dc=example, dc=com')
        self.assertEqual(dn[0], (('ou', 'dept1'),))
        self.assertEqual(dn[1], (('dc', 'example'),))
        self.assertEqual(dn[2], (('dc', 'com'),))
        self.assertEqual(DNObj(dn[:]), dn)
        with self.assertRaises(IndexError):
            self.assertEqual(dn[3], ())
        self.assertEqual(
            reversed(DNObj.from_str('ou=dept1,dc=example,dc=com')),
            DNObj.from_str('dc=com,dc=example,ou=dept1')
        )
        self.assertEqual(
            DNObj.from_str('ou=dept1,dc=example,dc=com')[1:],
            DNObj.from_str('dc=example,dc=com')[:]
        )
        self.assertEqual(
            DNObj.from_str('ou=dept1,dc=example,dc=com')[:2],
            DNObj.from_str('ou=dept1,dc=example')[:]
        )
        self.assertEqual(
            DNObj.from_str('ou=dept1,dc=example,dc=com').slice(1, None),
            DNObj.from_str('dc=example,dc=com')
        )
        self.assertEqual(
            DNObj.from_str('ou=dept1,dc=example,dc=com').slice(None, 2),
            DNObj.from_str('ou=dept1,dc=example')
        )
        domain = 'sub3.sub2.sub1.example.com'
        dns_labels = domain.split('.')
        dno = DNObj.from_domain(domain)
        for i, dn_comp in enumerate(dno):
            self.assertEqual(dn_comp[0][0], 'dc')
            self.assertEqual(dn_comp[0][1], dns_labels[i])
            self.assertEqual(dno[i], dn_comp)

    def test006_domain(self):
        self.assertEqual(
            str(DNObj.from_domain('example.com')),
            'dc=example,dc=com'
        )
        self.assertEqual(
            DNObj.from_domain('example.com').domain(),
            'example.com'
        )
        self.assertEqual(
            DNObj.from_domain('ströder.de').domain(),
            'ströder.de'.encode('idna').decode('ascii')
        )
        self.assertEqual(
            DNObj.from_str('ou=dept1,  dc=example, dc=com').domain(only_dc=False),
            'example.com'
        )
        with self.assertRaises(ValueError):
            DNObj.from_str('ou=dept1,dc=example,dc=com').domain(only_dc=True)
        with self.assertRaises(ValueError):
            DNObj.from_str('ou=dept1,dc=example,dc=com+dc=net').domain()
        with self.assertRaises(ValueError):
            DNObj.from_str('ou=foo+dc=example,dc=com').domain()


    def test007_add(self):
        self.assertEqual(
            DNObj.from_domain('dept1')+DNObj.from_domain('example.com'),
            DNObj.from_domain('dept1.example.com'),
        )
        self.assertEqual(
            bytes(DNObj.from_domain('dept1')+DNObj.from_domain('example.com')),
            b'dc=dept1,dc=example,dc=com',
        )
        self.assertEqual(
            (DNObj.from_domain('dept1')+DNObj.from_domain('example.com')).domain(),
            'dept1.example.com',
        )

    def test008_match(self):
        dn_obj = DNObj.from_domain('example.com')
        self.assertEqual(
            dn_obj.match([DNObj.from_domain('example.com')]),
            DNObj.from_domain('example.com')
        )
        self.assertEqual(
            dn_obj.match([DNObj.from_domain('sub1.example.com')]),
            DNObj((()))
        )
        self.assertEqual(
            dn_obj.match([
                DNObj.from_domain('example.com'),
                DNObj.from_domain('example.net'),
            ]),
            DNObj.from_domain('example.com')
        )
        self.assertEqual(
            dn_obj.match([
                DNObj.from_str('CN=Michael, dc=example, dc=com'),
                DNObj.from_str('DC=example, DC=com'),
            ]),
            DNObj.from_domain('example.com'),
        )
        self.assertEqual(
            DNObj.from_str('cn=Person,dc=foo,dc=example,dc=com').match([
                DNObj.from_domain('foo.example.com'),
                DNObj.from_domain('bar.example.com'),
            ]),
            DNObj.from_domain('foo.example.com'),
        )
        # sub entry match
        dn_obj = DNObj.from_domain('sub1.example.com')
        self.assertEqual(
            dn_obj.match([
                DNObj.from_domain('sub2.example.com'),
                DNObj.from_domain('example.com'),
                DNObj.from_domain('example.net'),
            ]),
            DNObj.from_domain('example.com')
        )
        self.assertEqual(
            dn_obj.match([
                DNObj.from_domain('sub1.example.com'),
                DNObj.from_domain('example.com'),
            ]),
            DNObj.from_domain('sub1.example.com')
        )
        # no match
        dn_obj = DNObj.from_domain('example.net')
        self.assertEqual(
            dn_obj.match([
                DNObj.from_domain('example.com'),
            ]),
            DNObj((()))
        )
        self.assertEqual(
            dn_obj.match([
                DNObj.from_domain('sub2.example.net'),
                DNObj.from_domain('example.com'),
            ]),
            DNObj((()))
        )
        # empty
        self.assertEqual(
            DNObj((())).match([DNObj((()))]),
            DNObj((()))
        )

    def test009_parents(self):
        self.assertEqual(
            DNObj.from_domain('example.com').parents(),
            [DNObj.from_domain('com')]
        )
        self.assertEqual(
            DNObj.from_domain('sub1.example.com').parents(),
            [
                DNObj.from_domain('example.com'),
                DNObj.from_domain('com'),
            ]
        )

    def test010_equals(self):
        self.assertTrue(
            DNObj.from_str('cn=Michael, dc=example, dc=com')==DNObj.from_str('CN=Michael, DC=example, DC=com')
        )
        self.assertFalse(
            DNObj.from_str('uid=Michael, dc=example, dc=com')==DNObj.from_str('CN=Michael, DC=example, DC=com')
        )
        self.assertFalse(
            DNObj.from_str('cn=Michael, dc=example, dc=com')==DNObj.from_str('CN=Mike, DC=example, DC=com')
        )
        self.assertEqual(
            DNObj.from_str('cn=Foo + mail= foo@example.com + sn = Bar,dc=example,dc=com'),
            DNObj.from_str('mail=foo@example.com+sn=Bar+cn=Foo,dc=example,dc=com'),
        )
        self.assertNotEqual(
            DNObj.from_str('cn=Foo2 + mail= foo@example.com + sn = Bar,dc=example,dc=com'),
            DNObj.from_str('mail=foo@example.com+sn=Bar+cn=Foo,dc=example,dc=com'),
        )
        self.assertNotEqual(
            DNObj.from_str('cn=Foo + mail= foo@example.com + sn = Bar,dc=example,dc=com'),
            DNObj.from_str('mail2=foo@example.com+sn=Bar+cn=Foo,dc=example,dc=com'),
        )

    def test010_at_sanitizer(self):
        self.assertNotEqual(
            str(DNObj.from_str('cn=Michael, dc=example, dc=com')),
            str(DNObj.from_str('CN=Michael, DC=example, DC=com'))
        )
        self.assertEqual(
            str(DNObj.from_str('cn=Michael, dc=example, dc=com')),
            str(DNObj.from_str('CN=Michael, DC=example, DC=com', at_sanitizer=str.lower))
        )


if __name__ == '__main__':
    unittest.main()



# -*- coding: ascii -*-
"""
ldap0.controls.simple - classes for some very simple LDAP controls
"""

import struct

from . import RequestControl, ResponseControl, LDAPControl


__all__ = [
    'AuthorizationIdentityRequestControl',
    'AuthorizationIdentityResponseControl',
    'BooleanControl',
    'ManageDSAITControl',
    'OctetStringInteger',
    'ProxyAuthzControl',
    'RelaxRulesControl',
    'ValueLessRequestControl',
]


class ValueLessRequestControl(RequestControl):
    """
    Base class for controls without a controlValue.
    The presence of the control in a LDAPv3 request changes the server's
    behaviour when processing the request simply based on the controlType.

    controlType
      OID of the request control
    criticality
      criticality request control
    """

    def __init__(self, controlType=None, criticality=False):
        RequestControl.__init__(
            self,
            controlType=controlType,
            criticality=criticality,
        )

    def encode(self) -> bytes:
        return None


class OctetStringInteger(LDAPControl):
    """
    Base class with controlValue being unsigend integer values

    integerValue
      Integer to be sent as OctetString
    """

    def __init__(self, controlType=None, criticality: bool = False, integerValue=None):
        LDAPControl.__init__(
            self,
            controlType=controlType,
            criticality=criticality,
        )
        self.integerValue = integerValue

    def encode(self) -> bytes:
        return struct.pack('!Q', self.integerValue)

    def decode(self, encodedControlValue: bytes):
        self.integerValue = struct.unpack('!Q', encodedControlValue)[0]


class BooleanControl(LDAPControl):
    """
    Base class for simple request controls with boolean control value.

    Constructor argument and class attribute:

    booleanValue
      Boolean (True/False or 1/0) which is the boolean controlValue.
    """
    boolean2ber = {
        1: b'\x01\x01\xFF',
        0: b'\x01\x01\x00',
    }
    ber2boolean = {
        b'\x01\x01\xFF': 1,
        b'\x01\x01\x00': 0,
    }

    def __init__(self, controlType=None, criticality: bool = False, booleanValue=False):
        LDAPControl.__init__(
            self,
            controlType=controlType,
            criticality=criticality,
        )
        self.booleanValue = booleanValue

    def encode(self) -> bytes:
        return self.boolean2ber[int(self.booleanValue)]

    def decode(self, encodedControlValue: bytes):
        self.booleanValue = bool(self.ber2boolean[encodedControlValue])


class ManageDSAITControl(ValueLessRequestControl):
    """
    Manage DSA IT Control
    """
    controlType: str = '2.16.840.1.113730.3.4.2'

    def __init__(self, criticality=False):
        ValueLessRequestControl.__init__(
            self,
            criticality=criticality,
        )


class RelaxRulesControl(ValueLessRequestControl):
    """
    Relax Rules Control
    """
    controlType: str = '1.3.6.1.4.1.4203.666.5.12'

    def __init__(self, criticality=False):
        ValueLessRequestControl.__init__(
            self,
            criticality=criticality,
        )


class ProxyAuthzControl(RequestControl):
    """
    Proxy Authorization Control

    authzId
      string containing the authorization ID indicating the identity
      on behalf which the server should process the request
    """
    controlType: str = '2.16.840.1.113730.3.4.18'

    def __init__(self, criticality: bool = False, authzId=''):
        RequestControl.__init__(
            self,
            criticality=criticality,
            encodedControlValue=authzId.encode(self.encoding),
        )


class AuthorizationIdentityRequestControl(ValueLessRequestControl):
    """
    Authorization Identity Request and Response Controls
    """
    controlType: str = '2.16.840.1.113730.3.4.16'

    def __init__(self, criticality=False):
        ValueLessRequestControl.__init__(
            self,
            criticality=criticality,
        )


class AuthorizationIdentityResponseControl(ResponseControl):
    """
    Authorization Identity Request and Response Controls

    Class attributes:

    authzId
        decoded authorization identity
    """
    controlType: str = '2.16.840.1.113730.3.4.15'

    def decode(self, encodedControlValue: bytes):
        self.authzId = encodedControlValue


class SubentriesControl(BooleanControl):
    """
    Subentries Control (RFC 3672)
    https://tools.ietf.org/html/rfc3672#section-3
    """
    controlType: str = '1.3.6.1.4.1.4203.1.10.1'

    def __init__(self, criticality: bool = False, booleanValue=False):
        BooleanControl.__init__(
            self,
            controlType=None,
            criticality=criticality,
            booleanValue=booleanValue
        )



# -*- coding: ascii -*-
"""
ldap0.controls.deref - classes for
(see https://tools.ietf.org/html/draft-masarati-ldap-deref)
"""

from pyasn1.type import namedtype, univ, tag
from pyasn1.codec.ber import encoder, decoder
from pyasn1.error import PyAsn1Error

from pyasn1_modules.rfc2251 import (
    LDAPDN,
    AttributeDescription,
    AttributeDescriptionList,
    AttributeValue,
)

from . import LDAPControl
from ..res import SearchResultEntry

__all__ = [
    'DereferenceControl',
]

# Request types
#---------------------------------------------------------------------------

# For compability with ASN.1 declaration in I-D
AttributeList = AttributeDescriptionList


class DerefSpec(univ.Sequence):
    componentType = namedtype.NamedTypes(
        namedtype.NamedType('derefAttr', AttributeDescription()),
        namedtype.NamedType('attributes', AttributeList()),
    )


class DerefSpecs(univ.SequenceOf):
    componentType = DerefSpec()


# Response types
#---------------------------------------------------------------------------


class AttributeValues(univ.SetOf):
    componentType = AttributeValue()


class PartialAttribute(univ.Sequence):
    componentType = namedtype.NamedTypes(
        namedtype.NamedType('type', AttributeDescription()),
        namedtype.NamedType('vals', AttributeValues()),
    )


class PartialAttributeList(univ.SequenceOf):
    componentType = PartialAttribute()
    tagSet = univ.Sequence.tagSet.tagImplicitly(
        tag.Tag(tag.tagClassContext, tag.tagFormatConstructed, 0)
    )


class DerefRes(univ.Sequence):
    componentType = namedtype.NamedTypes(
        namedtype.NamedType('derefAttr', AttributeDescription()),
        namedtype.NamedType('derefVal', LDAPDN()),
        namedtype.OptionalNamedType('attrVals', PartialAttributeList()),
    )


class DerefResultControlValue(univ.SequenceOf):
    componentType = DerefRes()


class DereferenceControl(LDAPControl):
    controlType: str = '1.3.6.1.4.1.4203.666.5.16'

    def __init__(self, criticality=False, derefSpecs=None):
        LDAPControl.__init__(self, criticality=criticality)
        self.derefSpecs = derefSpecs or {}

    def _deref_specs(self):
        deref_specs = DerefSpecs()
        i = 0
        for deref_attr, deref_attribute_names in self.derefSpecs.items():
            deref_spec = DerefSpec()
            deref_attributes = AttributeList()
            for j, deref_attribute_name in enumerate(deref_attribute_names):
                deref_attributes.setComponentByPosition(
                    j,
                    deref_attribute_name.encode(self.encoding)
                )
            deref_spec.setComponentByName(
                'derefAttr',
                AttributeDescription(deref_attr.encode(self.encoding))
            )
            deref_spec.setComponentByName('attributes', deref_attributes)
            deref_specs.setComponentByPosition(i, deref_spec)
            i += 1
        return deref_specs

    def encode(self) -> bytes:
        return encoder.encode(self._deref_specs())

    def decode(self, encodedControlValue: bytes):
        decoded_value, _ = decoder.decode(
            encodedControlValue,
            asn1Spec=DerefResultControlValue()
        )
        self.derefRes = {}
        for deref_res in decoded_value:
            deref_attr = bytes(deref_res[0])
            deref_dn = bytes(deref_res[1])
            deref_vals = deref_res[2]
            deref_res_entry_b = {}
            try:
                for tv in deref_vals or []:
                    deref_res_entry_b[bytes(tv[0])] = [bytes(tv1) for tv1 in tv[1]]
            except PyAsn1Error:
                pass
            deref_res_entry = SearchResultEntry(deref_dn, deref_res_entry_b, None)
            try:
                self.derefRes[deref_attr.decode(self.encoding)].append(deref_res_entry)
            except KeyError:
                self.derefRes[deref_attr.decode(self.encoding)] = [deref_res_entry]



# -*- coding: utf-8 -*-
"""
Automatic tests for module ldap0.base
"""

# from Python's standard lib
import unittest

from ldap0.base import decode_list, encode_list, decode_entry_dict, encode_entry_dict


class Test001Functions(unittest.TestCase):

    def test001_decode_list(self):
          self.assertEqual(
              decode_list(
                  [b'foo', b'Theo T\xc3\xa4ster', b'T\xc3\xa4ster, Theo'],
              ),
              ['foo', 'Theo Täster', 'Täster, Theo'],
          )

    def test002_decode_entry_dict(self):
          self.assertEqual(
              decode_entry_dict({
                    b'objectClass': [b'top', b'person', b'organizationalPerson', b'inetOrgPerson'],
                    b'cn': [b'Theo T\xc3\xa4ster', b'T\xc3\xa4ster, Theo'],
                    b'sn': [b'T\xc3\xa4ster'],
                    b'inv\xc3\xa4lid': [],
              }),
              {
                    'objectClass': ['top', 'person', 'organizationalPerson', 'inetOrgPerson'],
                    'cn': ['Theo Täster', 'Täster, Theo'],
                    'sn': ['Täster'],
                    'invälid': [],
              },
          )
          with self.assertRaises(UnicodeDecodeError):
              decode_entry_dict(
                  {
                        b'objectClass': [b'top', b'person', b'organizationalPerson', b'inetOrgPerson'],
                        b'cn': [b'Theo T\xc3\xa4ster', b'T\xc3\xa4ster, Theo'],
                        b'sn': [b'T\xc3\xa4ster'],
                        b'inv\xc3\xa4lid': [],
                  },
                  'ascii'
              )

    def test003_encode_list(self):
          self.assertEqual(
              encode_list(
                  ['foo', 'Theo Täster', 'Täster, Theo'],
              ),
              [b'foo', b'Theo T\xc3\xa4ster', b'T\xc3\xa4ster, Theo'],
          )

    def test004_encode_entry_dict(self):
          self.assertEqual(
              encode_entry_dict({
                    'objectClass': ['top', 'person', 'organizationalPerson', 'inetOrgPerson'],
                    'cn': ['Theo Täster', 'Täster, Theo'],
                    'sn': ['Täster'],
                    'invälid': [],
              }),
              {
                    b'objectClass': [b'top', b'person', b'organizationalPerson', b'inetOrgPerson'],
                    b'cn': [b'Theo T\xc3\xa4ster', b'T\xc3\xa4ster, Theo'],
                    b'sn': [b'T\xc3\xa4ster'],
                    b'inv\xc3\xa4lid': [],
              },
          )
          with self.assertRaises(UnicodeEncodeError):
              encode_entry_dict(
                  {
                        'objectClass': ['top', 'person', 'organizationalPerson', 'inetOrgPerson'],
                        'cn': ['Theo Täster', 'Täster, Theo'],
                        'sn': ['Täster'],
                        'invälid': [],
                  },
                  'ascii'
              )


if __name__ == '__main__':
    unittest.main()



# -*- coding: ascii -*-
"""
ldap0.ldif - generate and parse LDIF data (see RFC 2849)
"""

import re
from binascii import a2b_base64, b2a_base64
from binascii import Error as binascii_Error
from urllib.request import urlopen
from urllib.parse import urlparse
from io import BytesIO
from collections import defaultdict
from typing import BinaryIO, Dict, Optional, Sequence, Set, Tuple, Union

import _libldap0
from _libldap0 import MOD_ADD, MOD_DELETE, MOD_REPLACE, MOD_INCREMENT

from .functions import _libldap0_function_call
from .typehints import EntryBytes

__all__ = [
    'LDIFWriter',
    'LDIFParser',
]

MOD_OP_INTEGER = {
    b'add': MOD_ADD,
    b'delete': MOD_DELETE,
    b'replace': MOD_REPLACE,
    b'increment': MOD_INCREMENT,
}

MOD_OP_STR = {
    MOD_ADD: b'add',
    MOD_DELETE: b'delete',
    MOD_REPLACE: b'replace',
    MOD_INCREMENT: b'increment',
}

CHANGE_TYPES = {b'add', b'delete', b'modify', b'modrdn'}

SAFE_STRING_PATTERN = b'(^(\000|\n|\r| |:|<)|[\000\n\r\200-\377]+|[ ]+$)'
SAFE_STRING_RE = re.compile(SAFE_STRING_PATTERN)


class LDIFWriter:
    """
    Write LDIF entry or change records to file object
    Copy LDIF input to a file output object containing all data retrieved
    via URLs
    """
    _output_file: BinaryIO
    _base64_attrs: Set
    _cols: int
    _last_line_sep: bytes
    records_written: int

    def __init__(
            self,
            output_file: BinaryIO,
            base64_attrs: Optional[Sequence[bytes]] = None,
            cols: int = 76,
            line_sep=b'\n',
        ):
        """
        output_file
            file object for output
        base64_attrs
            list of attribute types to be base64-encoded in any case
        cols
            Specifies how many columns a line may have before it's
            folded into many lines.
        line_sep
            String used as line separator
        """
        self._output_file = output_file
        base64_attrs = base64_attrs or []
        self._base64_attrs = {a.lower() for a in base64_attrs}
        self._cols = cols
        self._last_line_sep = line_sep
        self.records_written = 0

    def _unfold_lines(self, line: bytes):
        """
        Write string line as one or more folded lines
        """
        # Check maximum line length
        line_len = len(line)
        if line_len <= self._cols:
            self._output_file.write(line)
            self._output_file.write(self._last_line_sep)
        else:
            # Fold line
            pos = self._cols
            self._output_file.write(line[0:min(line_len, self._cols)])
            self._output_file.write(self._last_line_sep)
            while pos < line_len:
                self._output_file.write(b' ')
                self._output_file.write(line[pos:min(line_len, pos+self._cols-1)])
                self._output_file.write(self._last_line_sep)
                pos = pos+self._cols-1
        # _unfold_lines()

    def _needs_base64_encoding(self, attr_type: bytes, attr_value: bytes) -> bool:
        """
        returns True if attr_value has to be base-64 encoded because
        of special chars or because attr_type is in self._base64_attrs
        """
        return (
            attr_type.lower() in self._base64_attrs or
            SAFE_STRING_RE.search(attr_value) is not None
        )

    def _unparse_attr_type_and_value(self, attr_type: bytes, attr_value: bytes):
        """
        Write a single attribute type/value pair

        attr_type
              attribute type
        attr_value
              attribute value
        """
        assert isinstance(attr_type, bytes), TypeError(
            'Expected bytes for attr_type, got %r' % (attr_type,)
        )
        assert isinstance(attr_value, bytes), TypeError(
            'Expected bytes for attr_value, got %r' % (attr_value,)
        )
        if self._needs_base64_encoding(attr_type, attr_value):
            # Encode with base64
            aval = b2a_base64(attr_value, newline=False)
            sep = b':: '
        else:
            aval = attr_value
            sep = b': '
        self._unfold_lines(sep.join((attr_type, aval)))
        # _unparseAttrTypeandValue()

    def _unparse_entry_record(self, entry: EntryBytes):
        """
        entry
            dictionary holding an entry
        """
        for attr_type in sorted(entry.keys()):
            for attr_value in entry[attr_type]:
                self._unparse_attr_type_and_value(attr_type, attr_value)

    def _unparse_change_record(self, modlist):
        """
        modlist
            list of additions (2-tuple) or modifications (3-tuple)
        """
        self._unparse_attr_type_and_value(b'changetype', b'modify')
        for mod in modlist:
            mod_op, mod_type, mod_vals = mod
            self._unparse_attr_type_and_value(MOD_OP_STR[mod_op], mod_type)
            if mod_vals:
                for mod_val in mod_vals:
                    self._unparse_attr_type_and_value(mod_type, mod_val)
            self._output_file.write(b'-'+self._last_line_sep)

    def unparse(self, dn, record):
        """
        dn
              string-representation of distinguished name
        record
              Either a dictionary holding the LDAP entry {attrtype:record}
              or a list with a modify list like for LDAPObject.modify().
        """
        # Start with line containing the distinguished name
        self._unparse_attr_type_and_value(b'dn', dn)
        # Dispatch to record type specific writers
        if isinstance(record, dict):
            self._unparse_entry_record(record)
        elif isinstance(record, list):
            self._unparse_change_record(record)
        else:
            raise ValueError('Argument record must be dictionary or list, was %r' % (record))
        # Write empty line separating the records
        self._output_file.write(self._last_line_sep)
        # Count records written
        self.records_written = self.records_written+1
        # unparse()


class LDIFParser:
    """
    Base class for a LDIF parser. Applications should sub-class this
    class and override method handle() to implement something meaningful.

    Public class attributes:

    records_read
          Counter for records processed so far
    """
    _input_file: BinaryIO
    _max_entries: Optional[int]
    _process_url_schemes: Set[bytes]
    _ignored_attr_types: Set[bytes]
    _last_line_sep: bytes
    version: Optional[int]
    line_counter: int
    byte_counter: int
    records_read: int
    changetype_counter: Dict[bytes, int]
    _last_line: Optional[bytes]

    def __init__(
            self,
            input_file: BinaryIO,
            ignored_attr_types: Optional[Sequence[bytes]] = None,
            max_entries: Optional[int] = None,
            process_url_schemes: Optional[Sequence[bytes]] = None,
            line_sep: bytes = b'\n'
        ):
        """
        Parameters:
        input_file
            File-object to read the LDIF input from
        ignored_attr_types
            Attributes with these attribute type names will be ignored.
        max_entries
            If non-zero specifies the maximum number of entries to be
            read from f.
        process_url_schemes
            List containing strings with URLs schemes to process with urllib.
            An empty list turns off all URL processing and the attribute
            is ignored completely.
        line_sep
            String used as line separator
        """
        self._input_file = input_file
        self._max_entries = max_entries
        self._process_url_schemes = {val.lower() for val in process_url_schemes or []}
        self._ignored_attr_types = {val.lower() for val in ignored_attr_types or []}
        self._last_line_sep = line_sep
        self.version = None
        # Initialize counters
        self.line_counter = 0
        self.byte_counter = 0
        self.records_read = 0
        self.changetype_counter = defaultdict(lambda: 0)
        # Read very first line
        try:
            self._last_line = self._readline()
        except EOFError:
            self._last_line = b''

    @classmethod
    def frombuf(cls, ldif_bytes: bytes, **kwargs):
        """
        initialize LDIFParser instance for parsing a byte sequence
        """
        assert isinstance(ldif_bytes, bytes), TypeError(
            'Expected bytes for ldif_bytes, got %r' % (ldif_bytes,)
        )
        return cls(BytesIO(ldif_bytes), **kwargs)

    def _readline(self) -> Union[bytes, None]:
        line = self._input_file.readline()
        self.line_counter = self.line_counter + 1
        self.byte_counter = self.byte_counter + len(line)
        if not line:
            return None
        if line[-2:] == b'\r\n':
            return line[:-2]
        if line[-1:] == b'\n':
            return line[:-1]
        return line

    def _unfold_lines(self) -> bytes:
        """
        Unfold several folded lines with trailing space into one line
        """
        if self._last_line is None:
            raise EOFError('EOF reached after %d lines (%d bytes)' % (
                self.line_counter,
                self.byte_counter,
            ))
        unfolded_lines = [self._last_line]
        next_line = self._readline()
        while next_line and next_line[0:1] == b' ':
            unfolded_lines.append(next_line[1:])
            next_line = self._readline()
        self._last_line = next_line
        return b''.join(unfolded_lines)

    def _next_key_and_value(self) -> Tuple[Union[bytes, None], Union[bytes, None]]:
        """
        Parse a single attribute type and value pair from one or
        more lines of LDIF data
        """
        attr_value: Union[bytes, None]
        # Reading new attribute line
        unfolded_line = self._unfold_lines()
        # Ignore comments which can also be folded
        while unfolded_line and unfolded_line[0:1] == b'#':
            unfolded_line = self._unfold_lines()
        if not unfolded_line:
            return None, None
        if unfolded_line == b'-':
            return b'-', None
        try:
            colon_pos = unfolded_line.index(b':')
        except ValueError:
            raise ValueError('no value-spec in %r' % (unfolded_line,))
        attr_type = unfolded_line[0:colon_pos]
        # if needed attribute value is BASE64 decoded
        value_spec = unfolded_line[colon_pos:colon_pos+2]
        if value_spec == b': ':
            attr_value = unfolded_line[colon_pos+2:].lstrip()
        elif value_spec == b'::':
            # attribute value needs base64-decoding
            try:
                attr_value = a2b_base64(unfolded_line[colon_pos+2:])
            except binascii_Error as err:
                raise ValueError('Error decoding base64 in %r: %s' % (unfolded_line, err))
        elif value_spec == b':<':
            # fetch attribute value from URL
            url = unfolded_line[colon_pos+2:].strip()
            attr_value = None
            if (
                    self._process_url_schemes and
                    urlparse(url)[0] in self._process_url_schemes
                ):
                attr_value = urlopen(url.decode('utf-8')).read()
        else:
            attr_value = unfolded_line[colon_pos+1:]
        return attr_type, attr_value

    def _consume_empty_lines(self) -> Tuple[Union[bytes, None], Union[bytes, None]]:
        """
        Consume empty lines until first non-empty line.
        Must only be used between full records!

        Returns non-empty key-value-tuple.
        """
        # Local symbol for better performance
        next_key_and_value = self._next_key_and_value
        # Consume empty lines
        try:
            key, val = next_key_and_value()
            while key == val == None:
                key, val = next_key_and_value()
        except EOFError:
            key, val = None, None
        return key, val

    def parse_entry_records(self, max_entries=None):
        """
        parse LDIF entry records
        """
        # Local symbol for better performance
        next_key_and_value = self._next_key_and_value
        if max_entries is None:
            max_entries = self._max_entries

        try:
            # Consume empty lines
            key, val = self._consume_empty_lines()
            # Consume version line
            if key == b'version':
                self.version = int(val)
                key, val = self._consume_empty_lines()
        except EOFError:
            return

        # Loop for processing whole records
        while key is not None and \
              (max_entries is None or self.records_read < max_entries):
            # Consume first line which must start with "dn: "
            if key != b'dn':
                raise ValueError(
                    'Line %d: First line of record does not start with "dn:": %r' % (
                        self.line_counter,
                        key,
                    )
                )
            try:
                _libldap0_function_call(_libldap0.str2dn, val, 1)
            except Exception:
                raise ValueError(
                    'Line %d: Not a valid string-representation for dn: %r' % (
                        self.line_counter,
                        val,
                    )
                )
            dn = val
            entry = {}
            # Consume second line of record
            key, val = next_key_and_value()

            # Loop for reading the attributes
            while key is not None:
                # Add the attribute to the entry if not ignored attribute
                if not key.lower() in self._ignored_attr_types:
                    try:
                        entry[key].append(val)
                    except KeyError:
                        entry[key] = [val]
                # Read the next line within the record
                try:
                    key, val = next_key_and_value()
                except EOFError:
                    key, val = None, None

            # handle record
            yield (dn, entry)
            self.records_read = self.records_read + 1
            # Consume empty separator line(s)
            key, val = self._consume_empty_lines()
        # parse_entry_records()

    def parse(self, max_entries=None):
        """
        Invokes LDIFParser.parse_entry_records() for backward compability
        """
        return self.parse_entry_records(max_entries=max_entries)

    def parse_change_records(self, max_entries=None):
        """
        parse LDIF change records
        """
        # Local symbol for better performance
        next_key_and_value = self._next_key_and_value
        if max_entries is None:
            max_entries = self._max_entries

        # Consume empty lines
        key, val = self._consume_empty_lines()
        # Consume version line
        if key == b'version':
            self.version = int(val)
            key, val = self._consume_empty_lines()

        # Loop for processing whole records
        while key is not None and \
              (max_entries is None or self.records_read < max_entries):
            # Consume first line which must start with "dn: "
            if key != b'dn':
                raise ValueError(
                    'Line %d: First line of record does not start with "dn:": %r' % (
                        self.line_counter,
                        key,
                    )
                )
            try:
                _libldap0_function_call(_libldap0.str2dn, val, 0)
            except Exception:
                raise ValueError(
                    'Line %d: Not a valid string-representation for dn: %r' % (
                        self.line_counter,
                        val,
                    )
                )
            dn = val
            # Consume second line of record
            key, val = next_key_and_value()
            # Read "control:" lines
            controls = []
            while key is not None and key == b'control':
                try:
                    control_type, criticality, control_value = val.split(' ', 2)
                except ValueError:
                    control_value = None
                    control_type, criticality = val.split(' ', 1)
                controls.append((control_type, criticality, control_value))
                key, val = next_key_and_value()

            # Determine changetype first
            changetype = None
            # Consume changetype line of record
            if key == b'changetype':
                if val not in CHANGE_TYPES:
                    raise ValueError('Invalid changetype: %r' % val)
                changetype = val
                key, val = next_key_and_value()

            if changetype == b'modify':
                # From here we assume a change record is read with changetype: modify
                modops = []
                try:
                    # Loop for reading the list of modifications
                    while key is not None:
                        # Extract attribute mod-operation (add, delete, replace)
                        try:
                            modop = MOD_OP_INTEGER[key]
                        except KeyError:
                            raise ValueError(
                                'Line %d: Invalid mod-op string: %r' % (self.line_counter, key)
                            )
                        # we now have the attribute name to be modified
                        modattr = val
                        modvalues = []
                        try:
                            key, val = next_key_and_value()
                        except EOFError:
                            key, val = None, None
                        while key == modattr:
                            modvalues.append(val)
                            try:
                                key, val = next_key_and_value()
                            except EOFError:
                                key, val = None, None
                        modops.append((modop, modattr, modvalues or None))
                        key, val = next_key_and_value()
                        if key == b'-':
                            # Consume next line
                            key, val = next_key_and_value()
                except EOFError:
                    key, val = None, None
                if modops:
                    # append entry to result list
                    yield (dn, modops, controls)
            else:
                # Consume the unhandled change record
                while key is not None:
                    key, val = next_key_and_value()

            # Consume empty separator line(s)
            key, val = self._consume_empty_lines()

            # Increment record counters
            try:
                self.changetype_counter[changetype] = self.changetype_counter[changetype] + 1
            except KeyError:
                self.changetype_counter[changetype] = 1
            self.records_read = self.records_read + 1

        # parse_change_records()

    def list_entry_records(self):
        """
        return list of parsed entry records
        """
        return list(self.parse_entry_records())

    def list_change_records(self):
        """
        return list of parsed change records
        """
        return list(self.parse_change_records())
